# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import abc

from . import LoopState


class StoppingCondition(abc.ABC):
    """ Chooses whether to stop the optimization based on the loop state """
    @abc.abstractmethod
    def should_stop(self, loop_state: LoopState) -> bool:
        """
        :param loop_state: Object that contains current state of the loop
        :return: Whether to stop collecting new data
        """
        pass


class FixedIterationsStoppingCondition(StoppingCondition):
    """ Stops after a fixed number of iterations """
    def __init__(self, i_max: int) -> None:
        """
        :param i_max: Maximum number of function observations within the loop, excluding initial points
        """
        self.i_max = i_max

    def should_stop(self, loop_state: LoopState) -> bool:
        """
        :param loop_state: Object that contains current state of the loop
        :return: True if maximum number of iterations has been reached
        """
        return loop_state.iteration >= self.i_max



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import Tuple

from ...core.acquisition import Acquisition
from .squared_correlation import SquaredCorrelation
from ..methods import VanillaBayesianQuadrature


class MutualInformation(Acquisition):
    """
    This acquisition function is the mutual information between the integral and the new point(s) under a GP-model.

    MutualInformation is a monotonic transform of SquaredCorrelation and hence yields the same acquisition policy in
    the case of vanilla-BQ.

    .. math::
        MI(x) = -0.5 \log(1-\rho^2(x))

    where :math:`\rho^2` is the SquaredCorrelation.
    """

    def __init__(self, model: VanillaBayesianQuadrature):
        """
        :param model: The vanilla Bayesian quadrature model
        """
        self.rho2 = SquaredCorrelation(model)

    def has_gradients(self) -> bool:
        return True

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the acquisition function at x.

        :param x: (n_points, input_dim) locations where to evaluate
        :return: (n_points, 1) the acquisition function value at x
        """
        rho2 = self.rho2.evaluate(x)
        mutual_information = -0.5 * np.log(1 - rho2)
        return mutual_information

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Evaluate the acquisition function with gradient

        :param x: (n_points, input_dim) locations where to evaluate
        :return: acquisition value and corresponding gradient at x, shapes (n_points, 1) and (n_points, input_dim)
        """
        # value
        mutual_information = self.evaluate(x)
        rho2, rho2_gradient = self.rho2.evaluate_with_gradients(x)

        # gradient
        mutual_information_gradient = (0.5 / (1 - rho2)) * rho2_gradient

        return mutual_information, mutual_information_gradient



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import Tuple, Union

from ...core.acquisition import Acquisition
from ...core.interfaces import IDifferentiable
from ..methods import WarpedBayesianQuadratureModel


class UncertaintySampling(Acquisition):
    """
    Uncertainty sampling acquisition function for (warped) Bayesian quadrature.

    The variance of the approximate transformed GP is used. If the integration measure is a probability measure,
    then the variance will be weighted with the probability density at each point.
    """

    def __init__(self, model: Union[WarpedBayesianQuadratureModel, IDifferentiable]):
        """
        :param model: A warped Bayesian quadrature model that has predictive gradients
        """
        self.model = model

    def has_gradients(self) -> bool:
        return True

    def _evaluate(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Evaluates the acquisition function. If a probability measure is used, then the variance will be weighted by
        the density of the probability measure.

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values
        :return: (n_points x 1) array of acquisition function values, unweighted variances
        """
        variances = self.model.predict(x)[1]
        if self.model.measure is None:
            return variances, variances
        else:
            weights = self.model.measure.compute_density(x).reshape(variances.shape)
            return variances * weights, variances

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the acquisition function. If a probability measure is used, then the variance will be weighted by
        the density of the probability measure.

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values
        :return: (n_points x 1) array of acquisition function values
        """
        return self._evaluate(x)[0]

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Optional abstract method that must be implemented if has_gradients returns True.
        Evaluates value and gradient of acquisition function at x.

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values and gradient
        :return: Tuple contains an (n_points x 1) array of acquisition function values and (n_points x n_dims) array of
                 acquisition function gradients with respect to x
        """
        variance_weighted, variance = self._evaluate(x)

        variance_gradient = self.model.get_prediction_gradients(x)[1]
        if self.model.measure is None:
            return variance, variance_gradient
        else:
            density = self.model.measure.compute_density(x)
            density_gradient = self.model.measure.compute_density_gradient(x)
            gradient_weighted = (density * variance_gradient.T).T + (variance[:, 0] * density_gradient.T).T
            return variance_weighted, gradient_weighted



import numpy as np
import scipy.optimize

from ..core import ParameterSpace
from ..core.acquisition import Acquisition
from ..core.interfaces import IDifferentiable
from ..core.loop import CandidatePointCalculator, LoopState
from ..bayesian_optimization.acquisitions.local_penalization import LocalPenalization
from ..core.optimization import AcquisitionOptimizerBase

N_SAMPLES = 500  # Number of samples to use when estimating Lipschitz constant
MAX_ITER = 200  # Maximum number of iterations for optimizer when estimating Lipschitz constant


class LocalPenalizationPointCalculator(CandidatePointCalculator):
    """
    Candidate point calculator that computes a batch using local penalization from:

    `Batch Bayesian Optimization via Local Penalization. Javier González, Zhenwen Dai, Philipp Hennig, Neil D. Lawrence
    <https://arxiv.org/abs/1505.08052>`_
    """

    def __init__(self, acquisition: Acquisition, acquisition_optimizer: AcquisitionOptimizerBase,
                 model: IDifferentiable, parameter_space: ParameterSpace, batch_size: int):
        """
        :param acquisition: Base acquisition function to use without any penalization applied, this acquisition should
                            output positive values only.
        :param acquisition_optimizer: AcquisitionOptimizer object to optimize the penalized acquisition
        :param model: Model object, used to compute the parameters of the local penalization
        :param parameter_space: Parameter space describing input domain
        :param batch_size: Number of points to collect in each batch
        """
        if not isinstance(model, IDifferentiable):
            raise ValueError('Model must implement ' + str(IDifferentiable) +
                             ' for use with Local Penalization batch method.')

        self.acquisition = acquisition
        self.acquisition_optimizer = acquisition_optimizer
        self.batch_size = batch_size
        self.model = model
        self.parameter_space = parameter_space

    def compute_next_points(self, loop_state: LoopState, context: dict=None) -> np.ndarray:
        """
        Computes a batch of points using local penalization.

        :param loop_state: Object containing the current state of the loop
        :param context: Contains variables to fix through optimization of acquisition function. The dictionary key is
                        the parameter name and the value is the value to fix the parameter to.
        """
        self.acquisition.update_parameters()

        # Initialize local penalization acquisition
        local_penalization_acquisition = LocalPenalization(self.model)

        # Everything done in log space so addition here is same as multiplying acquisition with local penalization
        # function.
        acquisition = self.acquisition + local_penalization_acquisition

        x_batch = []
        for i in range(self.batch_size):
            # Collect point
            x_next, _ = self.acquisition_optimizer.optimize(acquisition, context)
            x_batch.append(x_next)

            # Update local penalization acquisition with x_next
            f_min = np.min(self.model.Y)
            lipschitz_constant = _estimate_lipschitz_constant(self.parameter_space, self.model)
            local_penalization_acquisition.update_batches(np.concatenate(x_batch, axis=0), lipschitz_constant, f_min)
        return np.concatenate(x_batch, axis=0)


def _estimate_lipschitz_constant(space: ParameterSpace, model: IDifferentiable):
    """
    Estimate the lipschitz constant of the function by max norm of gradient currently in the model. Find this max
    gradient norm using an optimizer.
    """
    def negative_gradient_norm(x):
        d_mean_d_x, _ = model.get_prediction_gradients(x)
        result = np.sqrt((np.square(d_mean_d_x)).sum(1))  # simply take the norm of the expectation of the gradient
        return -result

    # Evaluate at some random samples first and start optimizer from point with highest gradient
    samples = space.sample_uniform(N_SAMPLES)
    samples = np.vstack([samples, model.X])
    gradient_norm_at_samples = negative_gradient_norm(samples)
    x0 = samples[np.argmin(gradient_norm_at_samples)][None, :]

    # Run optimizer to find point of highest gradient
    res = scipy.optimize.minimize(lambda x: negative_gradient_norm(x[None, :]), x0,
                                  bounds=space.get_bounds(),
                                  options={'maxiter': MAX_ITER})
    lipschitz_constant = -res.fun[0]

    min_lipschitz_constant = 1e-7
    fallback_lipschitz_constant = 10  # Value to use if calculated value is below minimum allowed
    if lipschitz_constant < min_lipschitz_constant:
        # To avoid problems in cases in which the model is flat.
        lipschitz_constant = fallback_lipschitz_constant
    return lipschitz_constant



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .loop_state import LoopState  # noqa: F401
from .user_function import UserFunction, UserFunctionWrapper  # noqa: F401
from .outer_loop import OuterLoop  # noqa: F401
from .user_function_result import UserFunctionResult  # noqa: F401
from .stopping_conditions import StoppingCondition, FixedIterationsStoppingCondition  # noqa: F401
from .model_updaters import ModelUpdater, FixedIntervalUpdater  # noqa: F401
from .candidate_point_calculators import CandidatePointCalculator, SequentialPointCalculator, RandomSampling # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import Union
from ...bayesian_optimization.acquisitions import ExpectedImprovement, ProbabilityOfFeasibility
from ...core.acquisition import Acquisition
from ...core.interfaces import IModel, IDifferentiable
from ...core.loop import FixedIntervalUpdater, OuterLoop, SequentialPointCalculator
from ...core.loop.loop_state import create_loop_state
from ...core.optimization import AcquisitionOptimizerBase
from ...core.optimization import GradientAcquisitionOptimizer
from ...core.parameter_space import ParameterSpace
from ..acquisitions.log_acquisition import LogAcquisition
from ..local_penalization_calculator import LocalPenalizationPointCalculator


class UnknownConstraintBayesianOptimizationLoop(OuterLoop):
    def __init__(self, space: ParameterSpace, model_objective: Union[IModel, IDifferentiable],
                 model_constraint: Union[IModel, IDifferentiable], acquisition: Acquisition = None,
                 update_interval: int = 1, batch_size: int = 1):

        """
        Emukit class that implements a loop for building Bayesian optimization with an unknown constraint.
        For more information see:

        Michael A. Gelbart, Jasper Snoek, and Ryan P. Adams,
        Bayesian Optimization with Unknown Constraints,
        https://arxiv.org/pdf/1403.5607.pdf

        :param space: Input space where the optimization is carried out.
        :param model_objective: The model that approximates the underlying objective function
        :param model_constraint: The model that approximates the unknown constraints
        :param acquisition: The acquisition function for the objective function (default, EI).
        :param update_interval:  Number of iterations between optimization of model hyper-parameters. Defaults to 1.
        :param batch_size: How many points to evaluate in one iteration of the optimization loop. Defaults to 1.
        """

        if not np.all(np.isclose(model_objective.X, model_constraint.X)):
            raise ValueError('Emukit currently only supports identical '
                             'training inputs for the constrained and objective model')

        if acquisition is None:
            acquisition = ExpectedImprovement(model_objective)
 
        acquisition_constraint = ProbabilityOfFeasibility(model_constraint)

        acquisition_constrained = acquisition * acquisition_constraint

        model_updater_objective = FixedIntervalUpdater(model_objective, update_interval)
        model_updater_constraint = FixedIntervalUpdater(model_constraint, update_interval,
                                                        lambda state: state.Y_constraint)

        acquisition_optimizer = GradientAcquisitionOptimizer(space)
        if batch_size == 1:
            candidate_point_calculator = SequentialPointCalculator(acquisition_constrained, acquisition_optimizer)
        else:
            log_acquisition = LogAcquisition(acquisition_constrained)
            candidate_point_calculator = LocalPenalizationPointCalculator(log_acquisition, acquisition_optimizer,
                                                                          model_objective, space, batch_size)

        loop_state = create_loop_state(model_objective.X, model_objective.Y, Y_constraint=model_constraint.Y)

        super(UnknownConstraintBayesianOptimizationLoop, self).__init__(candidate_point_calculator,
                                                                    [model_updater_objective, model_updater_constraint],
                                                                    loop_state)



import pickle
import numpy as np

from functools import partial
from typing import Tuple

try:
    import torch
    import torch.nn as nn
except ImportError:
    raise ImportError('pytorch is not installed. Please installed version it by running pip install torch torchvision')

try:
    from pybnn.util.layers import AppendLayer
except ImportError:
    raise ImportError('pybnn is not installed. Please install it by running pip install pybnn')

from emukit.core import ContinuousParameter, ParameterSpace
from emukit.core.loop.user_function import UserFunctionWrapper


def get_architecture_forrester(input_dimensionality: int) -> torch.nn.Module:
    class Architecture(nn.Module):
        def __init__(self, n_inputs, n_hidden=100):
            super(Architecture, self).__init__()
            self.fc1 = nn.Linear(n_inputs, n_hidden)
            self.fc2 = nn.Linear(n_hidden, 2)
            self.sigma_layer = AppendLayer(noise=1e-3)

        def forward(self, x):
            x = torch.tanh(self.fc1(x))
            x = self.fc2(x)
            mean = x[:, None, 0]
            return self.sigma_layer(mean)

    return Architecture(n_inputs=input_dimensionality)


def meta_forrester(fname_objective: str) -> Tuple[UserFunctionWrapper, ParameterSpace]:
    """
    Interface to the Meta-Forrester benchmark.

    Offline generated function samples can be download here:

    http://www.ml4aad.org/wp-content/uploads/2019/05/profet_data.tar.gz

    For further information about Profet and the generated meta-surrogate benchmarks see:

    Meta-Surrogate Benchmarking for Hyperparameter Optimization
    A. Klein and Z. Dai and F. Hutter and N. Lawrence and J. Gonzalez
    arXiv:1905.12982 [cs.LG] (2019)

    :param fname_objective: filename for the objective function
    :return: Tuple of user function object and parameter space
    """
    parameter_space = ParameterSpace([
        ContinuousParameter('x', 0, 1)])
    data = pickle.load(open(fname_objective, "rb"))
    x_mean_objective = data["x_mean"]
    x_std_objective = data["x_std"]
    y_mean_objective = data["y_mean"]
    y_std_objective = data["y_std"]
    task_feature_objective = data["task_feature"]

    objective = get_architecture_forrester(x_mean_objective.shape[0]).float()
    objective.load_state_dict(data["state_dict"])

    def objective_function(config):

        Ht = np.repeat(task_feature_objective[None, :], config.shape[0], axis=0)
        x = np.concatenate((config, Ht), axis=1)
        x_norm = torch.from_numpy((x - x_mean_objective) / x_std_objective).float()
        output = objective.forward(x_norm).data.numpy()
        mean = output[:, 0]

        feval = mean * y_std_objective + y_mean_objective

        return feval[:, None]

    f = partial(objective_function)

    return f, parameter_space



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .gpy_model_wrappers import GPyModelWrapper, GPyMultiOutputWrapper  # noqa: F401
from .gpy_quadrature_wrappers import BaseGaussianProcessGPy, RBFGPy  # noqa: F401
from .simple_gp_model import SimpleGaussianProcessModel  # noqa: F401



from typing import Tuple

import numpy as np

from ...core.acquisition import Acquisition


class LogAcquisition(Acquisition):
    """
    Takes the log of an acquisition function.
    """

    def __init__(self, acquisition: Acquisition):
        """
        :param acquisition: Base acquisition function that is log transformed. This acquisition function must output
                            positive values only.
        """
        self.acquisition = acquisition

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        :param x: Input location
        :return: log of original acquisition function at input location(s)
        """
        return np.log(self.acquisition.evaluate(x))

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Return log of original acquisition with gradient

        :param x: Input location
        :return: Tuple of (log value, gradient of log value)
        """
        value, gradient = self.acquisition.evaluate_with_gradients(x)

        epsilon = 1e-10
        value = np.maximum(value, epsilon)
        log_gradient = 1 / value * gradient
        return np.log(value), log_gradient

    @property
    def has_gradients(self) -> bool:
        return self.acquisition.has_gradients



from typing import Tuple

import numpy as np
from scipy.stats import norm

from ...core.acquisition import Acquisition
from ...core.interfaces import IModel


class LocalPenalization(Acquisition):
    """
    Penalization function used in the local penalization method. It is the log of the following function:

    .. math::
        h(x) = \sum_i{\log\\left(\Phi\\left(\\frac{\|x - x_i\|_{2} - r_i}{s_i}\\right)\\right)}

    where :math:`x_i` are the points already in the batch
    """
    def __init__(self, model: IModel):
        """
        :param model: Model
        """

        self.x_batch = None
        self.radius = None
        self.scale = None
        self.model = model

    @property
    def has_gradients(self) -> bool:
        return True

    def update_batches(self, x_batch: np.ndarray, lipschitz_constant: float, f_min: float):
        """
        Updates the batches internally and pre-computes the parameters of the penalization function
        """
        self.x_batch = x_batch
        if x_batch is not None:
            self._compute_parameters(x_batch, lipschitz_constant, f_min)
        else:
            self.x_batch = None
            self.radius = None
            self.scale = None

    def _compute_parameters(self, x_batch, lipschitz_constant, f_min):
        """
        Pre-computes the parameters of a penalization function
        """
        mean, variance = self.model.predict(x_batch)

        # Limit how small the variance can be to avoid numerical issues
        epsilon = 1e-10
        variance = np.maximum(epsilon, variance)

        std_deviation = np.sqrt(variance)

        # Calculate function parameters
        radius = (mean - f_min) / lipschitz_constant
        scale = std_deviation / lipschitz_constant
        self.radius = radius.flatten()
        self.scale = scale.flatten()
        self.x_batch = x_batch

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the penalization function value
        """

        if self.x_batch is None:
            return np.ones((x.shape[0], 1))

        distances = _distance_calculation(x, self.x_batch)
        normalized_distance = (distances - self.radius) / self.scale
        return norm.logcdf(normalized_distance).sum(axis=1, keepdims=True)

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Evaluates the penalization function value and gradients with respect to x
        """

        if self.x_batch is None:
            return np.ones((x.shape[0], 1)), np.zeros(x.shape)

        distances, d_dist_dx = _distance_with_gradient(x, self.x_batch)
        normalized_distance = (distances - self.radius) / self.scale
        h_func = norm.cdf(normalized_distance)
        d_value_dx = 0.5 * (1 / h_func[:, :, None]) \
                     * norm.pdf(normalized_distance)[:, :, None] \
                     * d_dist_dx / self.scale[None, :, None]
        return norm.logcdf(normalized_distance).sum(1, keepdims=True), d_value_dx.sum(1)


def _distance_calculation(x_1, x_2):
    dx = x_1[:, None, :] - x_2[None, :, :]
    return np.sqrt(np.square(dx).sum(-1))


def _distance_with_gradient(x_1, x_2):
    distances = _distance_calculation(x_1, x_2)
    inv_distance = np.where(distances != 0., 1 / distances, 0)
    dx = x_1[:, None, :] - x_2[None, :, :]
    d_dist_dx = 2 * dx * inv_distance[:, :, None]
    return distances, d_dist_dx



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Tuple

import numpy as np

from ...core import ContinuousParameter, InformationSourceParameter, ParameterSpace
from ...core.loop.user_function import MultiSourceFunctionWrapper


def multi_fidelity_borehole_function(high_noise_std_deviation: float = 0, low_noise_std_deviation: float = 0) \
        -> Tuple[MultiSourceFunctionWrapper, ParameterSpace]:
    """
    Two level borehole function.

    The Borehole function models water flow through a borehole. Its simplicity and quick evaluation makes it a commonly
    used function for testing a wide variety of methods in computer experiments.

    See reference for equations:
    https://www.sfu.ca/~ssurjano/borehole.html

    :param high_noise_std_deviation: Standard deviation of Gaussian observation noise on high fidelity observations.
                                     Defaults to zero.
    :param low_noise_std_deviation: Standard deviation of Gaussian observation noise on low fidelity observations.
                                     Defaults to zero.
    :return: Tuple of user function object and parameter space
    """
    parameter_space = ParameterSpace([
        ContinuousParameter('borehole_radius', 0.05, 0.15),
        ContinuousParameter('radius_of_influence', 100, 50000),
        ContinuousParameter('upper_aquifer_transmissivity', 63070, 115600),
        ContinuousParameter('upper_aquifer_head', 990, 1110),
        ContinuousParameter('lower_aquifer_transmissivity', 63.1, 116),
        ContinuousParameter('lower_aquifer_head', 700, 820),
        ContinuousParameter('borehole_length', 1120, 1680),
        ContinuousParameter('hydraulic_conductivity', 9855, 12045),
        InformationSourceParameter(2)])

    user_function = MultiSourceFunctionWrapper([
        lambda x: _borehole_low(x, low_noise_std_deviation),
        lambda x: _borehole_high(x, high_noise_std_deviation)])

    return user_function, parameter_space


def _borehole_high(x, sd=0):
    """
    High fidelity version of borehole function

    Reference:
    https://www.sfu.ca/~ssurjano/borehole.html
    """

    numerator = 2 * np.pi * x[:, 2] * (x[:, 3] - x[:, 5])
    ln_r_rw = np.log(x[:, 1] / x[:, 0])
    denominator = ln_r_rw * \
        (x[:, 2] / x[:, 4] + 1 + (2 * x[:, 6] * x[:, 2]) /
         (x[:, 0]**2 * x[:, 7] * ln_r_rw))
    return (numerator / denominator)[:, None] + np.random.randn(x.shape[0], 1) * sd


def _borehole_low(x, sd=0):
    """
    Low fidelity version of borehole function

    Reference:
    https://www.sfu.ca/~ssurjano/borehole.html
    """

    numerator = 5 * x[:, 2] * (x[:, 3] - x[:, 5])
    ln_r_rw = np.log(x[:, 1] / x[:, 0])
    denominator = ln_r_rw * \
        (x[:, 2] / x[:, 4] + 1.5 + (2 * x[:, 6] * x[:, 2]) /
         (x[:, 0]**2 * x[:, 7] * ln_r_rw))
    return (numerator / denominator)[:, None] + np.random.randn(x.shape[0], 1) * sd



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Union, Callable

import scipy
import numpy as np
from scipy.stats import norm
from scipy.optimize import bisect

from ...core.acquisition import Acquisition
from ...core.interfaces import IModel
from ...core.parameter_space import ParameterSpace
from ...core.initial_designs import RandomDesign

from ..interfaces import IEntropySearchModel


class MaxValueEntropySearch(Acquisition):
    def __init__(self, model: Union[IModel, IEntropySearchModel], space: ParameterSpace,
                 num_samples: int = 10, grid_size: int = 5000) -> None:

        """
        MES acquisition function approximates the distribution of the value at the global
        minimum and tries to decrease its entropy. See this paper for more details:
        Z. Wang, S. Jegelka
        Max-value Entropy Search for Efficient Bayesian Optimization
        ICML 2017

        :param model: GP model to compute the distribution of the minimum dubbed pmin.
        :param space: Domain space which we need for the sampling of the representer points
        :param num_samples: integer determining how many samples to draw of the minimum (does not need to be large)
        :param grid_size: number of random locations in grid used to fit the gumbel distribution and approximately generate
        the samples of the minimum (recommend scaling with problem dimension, i.e. 10000*d)
        """
        super().__init__()

        if not isinstance(model, IEntropySearchModel):
            raise RuntimeError("Model is not supported for MES")

        self.model = model
        self.space = space
        self.num_samples = num_samples
        self.grid_size = grid_size

        # Initialize parameters to lazily compute them once needed
        self.mins = None

    def update_parameters(self):
        # apply gumbel sampling to obtain samples from y*
        # we approximate Pr(y*^hat<y) by Gumbel(alpha,beta)
        # generate grid
        N = self.model.model.X.shape[0]

        random_design = RandomDesign(self.space)
        grid = random_design.get_samples(self.grid_size)
        fmean, fvar = self.model.model.predict(np.vstack([self.model.model.X, grid]), include_likelihood=False)
        fsd = np.sqrt(fvar)
        idx = np.argmin(fmean[:N])

        # scaling so that gumbel scale is proportional to IQ range of cdf Pr(y*<z)
        # find quantiles Pr(y*<y1)=r1 and Pr(y*<y2)=r2
        right = fmean[idx].flatten()
        left = right
        probf = lambda x: np.exp(np.sum(norm.logcdf(-(x - fmean) / fsd), axis=0))
        i = 0
        while probf(left) < 0.75:
            left = 2. ** i * np.min(fmean - 5. * fsd) + (1. - 2. ** i) * right
            i += 1
        i = 0
        while probf(right) > 0.25:
            right = -2. ** i * np.min(fmean - 5. * fsd) + (1. + 2. ** i) * fmean[idx].flatten()
            i += 1

        # Binary search for 3 percentiles
        q1, med, q2 = map(lambda val: bisect(lambda x: probf(x) - val, left, right, maxiter=10000, xtol=0.00001),
                            [0.25, 0.5, 0.75])

        # solve for gumbel params
        beta = (q1 - q2) / (np.log(np.log(4. / 3.)) - np.log(np.log(4.)))
        alpha = med + beta * np.log(np.log(2.))

        # sample K length vector from unif([0,1])
        # return K Y* samples
        self.mins = -np.log(-np.log(np.random.rand(self.num_samples))) * beta + alpha

    def _required_parameters_initialized(self):
        """
        Checks if all required parameters are initialized.
        """
        return self.mins is not None

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the information gain, i.e the change in entropy of p_min if we would evaluate x.
        :param x: points where the acquisition is evaluated.
        """
        if not self._required_parameters_initialized():
            self.update_parameters()
        fmean, fvar = self.model.predict(x)
        fsd = np.sqrt(fvar)
        gamma = (self.mins - fmean) / fsd
        f_acqu_x = np.mean(-gamma * norm.pdf(gamma) / (2 * (1 - norm.cdf(gamma))) - np.log(1 - norm.cdf(gamma)),
                            axis=1)
        return f_acqu_x.reshape(-1, 1)

    @property
    def has_gradients(self) -> bool:
        """Returns that this acquisition has gradients"""
        return False



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import Tuple, List

from ...core.continuous_parameter import ContinuousParameter


class BoxBounds:
    """
    Box bounds
    """
    def __init__(self, name: str, bounds: List[Tuple[float, float]]):
        """
        :param name: Name of parameter
        :param bounds: List of D tuples, where D is the dimensionality and the tuples contain the
        lower and upper bounds of the box i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]
        """

        self.name = name
        self.dim = len(bounds)
        self._check_bound_validity(bounds)
        self._bounds = bounds
        self.lower_bounds, self.upper_bounds = self.get_lower_and_upper_bounds()

    @property
    def bounds(self):
        return self._bounds

    @bounds.setter
    def bounds(self, new_bounds: List[Tuple[float, float]]) -> None:
        """
        Sets new box bounds and checks their validity
        :param new_bounds: List of D tuples, where D is the dimensionality of the box and the tuples contain the
        lower and upper bounds of the box i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]
        """
        if not len(new_bounds) == self.dim:
            raise ValueError('Length of new box bounds is {} (length {} expected).'.format(len(new_bounds), self.dim))

        self._check_bound_validity(new_bounds)
        self._bounds = new_bounds
        self.lower_bounds, self.upper_bounds = self.get_lower_and_upper_bounds()

    def _check_bound_validity(self, bounds: List[Tuple[float, float]]) -> None:
        """
        checks if lower bounds are smaller than upper bounds.
        :param bounds: List of D tuples, where D is the dimensionality of the box and the tuples contain the
        lower and upper bounds of the box i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]
        """
        if self.dim == 0:
            raise ValueError("Length of bound list must be > 0; empty list found.")
        for bounds_d in bounds:
            lb_d, ub_d = bounds_d
            if lb_d >= ub_d:
                raise ValueError("Upper box bound must be larger than lower bound. Found a pair containing ({}, "
                                 "{}).".format(lb_d, ub_d))

    def get_lower_and_upper_bounds(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        returns two arrays, one containing all lower bounds and one containing all upper bounds.
        :return: lower bounds, upper bounds of box, shapes (1, input_dim)
        """
        lower_bounds = np.zeros([self.dim, 1])
        upper_bounds = np.zeros([self.dim, 1])
        for i, bounds_d in enumerate(self._bounds):
            lb_d, ub_d = bounds_d
            lower_bounds[i] = lb_d
            upper_bounds[i] = ub_d
        return lower_bounds.T, upper_bounds.T

    def convert_to_list_of_continuous_parameters(self) -> List[ContinuousParameter]:
        """
        converts the box bounds into a list of ContinuousParameter objects
        :return: a list if ContinuousParameter objects (one for each dimension)
        """
        continuous_parameters = []
        for i, bounds_d in enumerate(self._bounds):
            lb_d, ub_d = bounds_d
            name_d = self.name + '_' + str(i)
            param = ContinuousParameter(name=name_d, min_value=lb_d, max_value=ub_d)
            continuous_parameters.append(param)
        return continuous_parameters



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple, List
import numpy as np

from emukit.core.loop.user_function import UserFunctionWrapper


def hennig1D() -> Tuple[UserFunctionWrapper, List[Tuple[float, float]]]:
    """
    1D toy integrand coined by Philipp Hennig

    .. math::

        e^{-x^2 -\sin^2(3x)}

    :return: the wrapped test function, and the integrals bounds (defaults to interval [-3, 3]).
    """
    integral_bounds = [(-3., 3.)]
    return UserFunctionWrapper(_hennig1D), integral_bounds


def _hennig1D(x: np.ndarray) -> np.ndarray:
    """
    :param x: locations for evaluation (num_points, 1)
    :return: the function values at x, shape (num_points, 1)
    """
    return np.exp(- x**2 - np.sin(3. * x)**2)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


"""
Contains integrated variance acquisition
"""
import numpy as np

from ...core.acquisition import Acquisition
from ...core.initial_designs import RandomDesign
from ...core import ParameterSpace

from ..interfaces import ICalculateVarianceReduction


class IntegratedVarianceReduction(Acquisition):
    """
    Acquisition function for integrated variance reduction
    """

    def __init__(self,
                 model: ICalculateVarianceReduction, space: ParameterSpace,
                 x_monte_carlo=None, num_monte_carlo_points: int=int(1e5)) -> None:
        """
        :param model: The emulation model
        :param space: The parameter space to select points within
        :param x_monte_carlo: The points to evaluate the GP at in order to estimate the integral of the variance.
                              Can be None and we will randomly sample within the constraints defined by the space.
        :param num_monte_carlo_points: Number of points to use to do Monte Carlo integration of variance.
                                       Not used if x_monte_carlo supplied.
        """

        self.model = model

        if x_monte_carlo is None:
            # Use RandomDesign to generate random points to do Monte Carlo integration
            # while respecting any constraints
            random_design = RandomDesign(space)
            self._x_monte_carlo = random_design.get_samples(num_monte_carlo_points)
        else:
            # Use user supplied points
            in_domain = space.check_points_in_domain(x_monte_carlo)
            if not np.all(in_domain):
                raise ValueError('Some or all of the points in x_monte_carlo are out of the valid domain.')
            self._x_monte_carlo = x_monte_carlo

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        :param x: The new training point(s) to evaluate with shape (n_points x 1)
        :return: A numpy array with shape (n_points x 1)
                 containing the values of the acquisition evaluated at each x row
        """

        n_eval_points = x.shape[0]
        integrated_variance = np.zeros((n_eval_points, 1))
        for i in range(n_eval_points):
            # Find variance reduction at each Monte Carlo point
            variance_reduction = self.model.calculate_variance_reduction(x[[i], :], self._x_monte_carlo)
            # Take mean to approximate integral per unit volume
            integrated_variance[i] = np.mean(variance_reduction)

        return integrated_variance

    @property
    def has_gradients(self):
        return False



import logging
from typing import Callable, Optional

import numpy as np

_log = logging.getLogger(__name__)


class IConstraint:
    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        :param x: Array of shape (n_points x n_dims) containing input locations to evaluate constraint at
        :return: Numpy array of shape (n_input,) where an element will be 1 if the corresponding input satisfies the
                 constraint and zero if the constraint is violated
        """
        raise NotImplementedError


class InequalityConstraint(IConstraint):
    def __init__(self, lower_bound: np.ndarray, upper_bound: np.ndarray):
        """
        :param lower_bound: Lower bound vector of size (n_constraint,). Can be -np.inf for one sided constraint
        :param upper_bound: Upper bound vector of size (n_constraint,). Can be np.inf for one sided constraint
        """
        if (lower_bound is None) and (upper_bound is None):
            raise ValueError('Neither lower nor upper bounds is set, at least one must be specified')

        # Default lower bound to -infinity
        if lower_bound is None:
            lower_bound = np.full([upper_bound.shape[0]], -np.inf)

        # Default upper bound to +infinity
        if upper_bound is None:
            upper_bound = np.full([lower_bound.shape[0]], np.inf)

        if np.any((lower_bound == -np.inf) & (upper_bound == np.inf)):
            raise ValueError('One or more inequality constraints are unbounded')

        if np.any(lower_bound >= upper_bound):
            raise ValueError('Lower bound is greater than or equal to upper bound for one or more constraints')

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound


class LinearInequalityConstraint(InequalityConstraint):
    """
    Constraint of the form lower_bound < Ax < upper_bound where the matrix A is called "constraint_matrix"
    """
    def __init__(self, constraint_matrix: np.ndarray, lower_bound: np.ndarray=None, upper_bound: np.ndarray=None):
        """

        :param constraint_matrix: (n_constraint, n_x_dims) matrix in b_lower < Ax < b_upper
        :param lower_bound: Lower bound vector of size (n_constraint,). Can be -np.inf for one sided constraint
        :param upper_bound: Upper bound vector of size (n_constraint,). Can be np.inf for one sided constraint
        """
        super().__init__(lower_bound, upper_bound)
        self.constraint_matrix = constraint_matrix

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate whether constraints are violated or satisfied at a set of x locations

        :param x: Array of shape (n_points x n_dims) containing input locations to evaluate constraint at
        :return: Numpy array of shape (n_input,) where an element will be 1 if the corresponding input satisfies the
                 constraint and zero if the constraint is violated
        """
        ax = self.constraint_matrix.dot(x)
        return np.all([ax >= self.lower_bound, ax <= self.upper_bound], axis=0)


class NonlinearInequalityConstraint(InequalityConstraint):
    """
    Constraint of the form lower_bound < g(x) < upper_bound
    """
    def __init__(self, constraint_function: Callable, lower_bound: np.ndarray, upper_bound: np.ndarray,
                 jacobian_fun: Optional[Callable]=None):
        """
        :param constraint_function: function defining constraint in b_lower < fun(x) < b_upper.
                                    Has signature f(x) -> array, shape(m,) where x is 1d and m is the number of constraints
        :param lower_bound: Lower bound vector of size (n_constraint,). Can be -np.inf for one sided constraint
        :param upper_bound: Upper bound vector of size (n_constraint,). Can be np.inf for one sided constraint
        :param jacobian_fun: Function returning the jacobian of the constraint function. Optional, if not supplied
                             the optimizer will use finite differences to calculate the gradients of the constraint
        """

        super().__init__(lower_bound, upper_bound)

        self.fun = constraint_function
        self.jacobian_fun = jacobian_fun

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate whether constraints are violated or satisfied at a set of x locations

        :param x: Array of shape (n_points x n_dims) containing input locations to evaluate constraint at
        :return: Numpy array of shape (n_input,) where an element will be 1 if the corresponding input satisfies the
                 constraint and zero if the constraint is violated
        """
        fun_x = np.array([self.fun(x) for x in x])
        return np.all([fun_x >= self.lower_bound, fun_x <= self.upper_bound], axis=0)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from ...core.interfaces.models import IModel, IDifferentiable

try:
    from pybnn import bohamiann
except ImportError:
    raise ImportError("""
        This module is missing required dependencies. Try running

        pip install git+https://github.com/automl/pybnn.git

        Refer to https://github.com/automl/pybnn for further information.
    """)

import torch
import torch.nn as nn


def get_default_network(input_dimensionality: int) -> torch.nn.Module:
    class AppendLayer(nn.Module):
        def __init__(self, bias=True, *args, **kwargs):
            super().__init__(*args, **kwargs)
            if bias:
                self.bias = nn.Parameter(torch.Tensor(1, 1))
            else:
                self.register_parameter('bias', None)

        def forward(self, x):
            return torch.cat((x, self.bias * torch.ones_like(x)), dim=1)

    def init_weights(module):
        if type(module) == AppendLayer:
            nn.init.constant_(module.bias, val=np.log(1e-3))
        elif type(module) == nn.Linear:
            nn.init.kaiming_normal_(module.weight, mode="fan_in", nonlinearity="linear")
            nn.init.constant_(module.bias, val=0.0)

    return nn.Sequential(
        nn.Linear(input_dimensionality, 50), nn.Tanh(),
        nn.Linear(50, 50), nn.Tanh(),
        nn.Linear(50, 1),
        AppendLayer()
    ).apply(init_weights)


class Bohamiann(IModel, IDifferentiable):

    def __init__(self, X_init: np.ndarray, Y_init: np.ndarray, num_steps: int = 5000, num_burnin: int = 5000,
                 lr: float = 1e-2, get_architecture=get_default_network,
                 **kwargs) -> None:
        """
        Implements Bayesian neural networks as described by Springenberg et. al[1] based on
        stochastic gradient Hamiltonian monte carlo sampling[2].

        Dependencies:
            AutoML pybnn (https://github.com/automl/pybnn)

        [1] J. T. Springenberg, A. Klein, S. Falkner, F. Hutter
            Bayesian Optimization with Robust Bayesian Neural Networks.
            In Advances in Neural Information Processing Systems 29 (2016).
        [2] T. Chen, E. B. Fox, C. Guestrin
            Stochastic Gradient Hamiltonian Monte Carlo
            Proceedings of the 31st International Conference on Machine Learning

        :param X_init: training data points
        :param Y_init: training function values
        :param num_steps: number of MCMC steps (after the burnin)
        :param num_burnin: number of burnin steps
        :param lr: learning rate or step length of the MCMC sampler
        :param get_architecture: function handle that returns an architecture for the Bayesian neural network
        """
        super().__init__()

        self.model = bohamiann.Bohamiann(get_network=get_architecture)
        self.num_steps = num_steps
        self.num_burnin = num_burnin
        self._X = X_init
        self._Y = Y_init

        self.model.train(X_init, Y_init, num_steps=self.num_steps + self.num_burnin, lr=lr,
                         num_burn_in_steps=self.num_burnin, keep_every=100, **kwargs)

    @property
    def X(self):
        return self._X

    @property
    def Y(self):
        return self._Y

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Predict values for given points

        :param X: points to run prediction for
        """
        m, v = self.model.predict(X)

        return m[:, None], v[:, None]

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Sets training data in model

        :param X: new points
        :param Y: function values at new points X
        """
        self._X = X
        self._Y = Y

        self.model.train(self._X, self._Y, num_steps=self.num_steps,
                         num_burn_in_steps=self.num_burnin, keep_every=100, verbose=True)

    def optimize(self) -> None:
        pass

    def get_f_minimum(self):
        return np.min(self._Y)

    def get_prediction_gradients(self, X: np.ndarray) -> np.ndarray:
        """
        Computes and returns model gradients at given points

        :param X: points to compute gradients at
        """
        dm = np.array([self.model.predictive_mean_gradient(xi) for xi in X])
        dv = np.array([self.model.predictive_variance_gradient(xi) for xi in X])

        return dm, dv



import torch
import numpy as np


def objective_function(config, model_objective, model_cost,
                       task_feature_objective, task_feature_cost,
                       x_mean_objective, x_std_objective,
                       x_mean_cost, x_std_cost,
                       y_mean_objective=None, y_std_objective=None,
                       y_mean_cost=None, y_std_cost=None,
                       log_objective=False,
                       with_noise=True):

    Ht = np.repeat(task_feature_objective[None, :], config.shape[0], axis=0)
    x = np.concatenate((config, Ht), axis=1)
    x_norm = torch.from_numpy((x - x_mean_objective) / x_std_objective).float()
    output = model_objective.forward(x_norm).data.numpy()

    mean = output[:, 0]
    log_variance = output[:, 1]
    if y_mean_objective is not None or y_std_objective is not None:
        mean = mean * y_std_objective + y_mean_objective
        log_variance *= y_std_objective ** 2

    feval = mean
    if with_noise:
        feval += np.random.randn() * np.sqrt(np.exp(log_variance))

    if log_objective:
        feval = np.exp(feval)

    Ht = np.repeat(task_feature_cost[None, :], config.shape[0], axis=0)
    x = np.concatenate((config, Ht), axis=1)
    x_norm = torch.from_numpy((x - x_mean_cost) / x_std_cost).float()
    output = model_cost.forward(x_norm).data.numpy()

    log_mean = output[:, 0]
    log_log_variance = output[:, 1]
    if y_mean_cost is not None or y_std_cost is not None:
        log_mean = log_mean * y_std_cost + y_mean_cost
        log_log_variance *= y_std_cost ** 2

    log_cost = log_mean
    if with_noise:
        log_cost += np.random.randn() * np.sqrt(np.exp(log_log_variance))

    return feval[:, None], np.exp(log_cost)[:, None]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np


class IEntropySearchModel:
    """
    Interface containing abstract methods that need to be implemented if using entropy search Bayesian optimization
    acquisition function.
    """
    def predict_covariance(self, X: np.ndarray, with_noise: bool=True) -> np.ndarray:
        """

        :param X: Numpy array of shape (n_points, n_features) of test input locations
        :param with_noise: Whether to include likelihood noise term in covariance
        :return: Posterior covariance matrix which has shape (n_points x n_points)
        """
        raise NotImplementedError

    def get_covariance_between_points(self, X1: np.ndarray, X2: np.ndarray) -> np.ndarray:
        """
        Calculate posterior covariance between one point in X1 and all points in X2

        :param X1: An array of shape 1 x n_dimensions that contains a data single point. It is the first argument of the
                   posterior covariance function
        :param X2: An array of shape n_points x n_dimensions that may contain multiple data points. This is the second
                   argument to the posterior covariance function.
        :return: An array of shape n_points x 1 of posterior covariances between X1 and X2
        """
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .models import IModel, IDifferentiable, IPriorHyperparameters, IJointlyDifferentiable  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple, List
import numpy as np

from emukit.core.loop.user_function import UserFunctionWrapper


def sombrero2D(freq: float = 1.) -> Tuple[UserFunctionWrapper, List[Tuple[float, float]]]:
    """
    2D Sombrero function

    .. math::

        \frac{\sin(\pi r freq)}{\pi r freq}

    :param freq: frequency of the sombrero (must be > 0, defaults to 1)
    :return: the wrapped test function, and the integrals bounds (defaults to area [-3, 3]^2).
    """
    func = lambda x: _sombrero2D(x, freq)
    integral_bounds = 2 * [(-3., 3.)]
    return UserFunctionWrapper(func), integral_bounds


def _sombrero2D(x: np.ndarray, freq: float) -> np.ndarray:
    """
    :param x: locations for evaluation (num_points, 2)
    :param freq: frequency of the sombrero (must be > 0)
    :return: the function values at x, shape (num_points, 1)
    """
    r = np.sqrt((x*x).sum(axis=1))
    r_scaled = (np.pi * freq) * r
    result = np.sin(r_scaled) / r_scaled
    result[np.isnan(result)] = 1.
    return result[:, np.newaxis]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import Tuple, Callable

from ...core import ParameterSpace
from ...core.loop import UserFunctionWrapper
from ...core.interfaces.models import IModel


class ModelFreeMonteCarloSensitivity(object):
    """
    Class to do sensitivity analysis of a function. It computes Monte Carlo approximations to
    the Sobol indexes and the total variance components of each input variable of some objective
    of interest.
    """

    def __init__(self, objective: Callable, input_domain: ParameterSpace) -> None:
        """
        :param objective: python function in which the sensitivity analysis will be performed.
        :param input_domain: parameter space.
        """
        self.objective = UserFunctionWrapper(objective)
        self.input_domain = input_domain

    def _generate_samples(self, num_monte_carlo_points: int=int(1e5)) -> None:
        """
        Generates the two samples that are used to compute the main and total indices

        :param num_monte_carlo_points: number of samples to generate
        """
        self.main_sample = self.input_domain.sample_uniform(num_monte_carlo_points)
        self.fixing_sample = self.input_domain.sample_uniform(num_monte_carlo_points)

    def saltelli_estimators(self,
                            f_main_sample: np.ndarray, f_fixing_sample: np.ndarray,
                            f_new_fixing_sample: np.ndarray, num_monte_carlo_points: int,
                            total_mean: np.float64, total_variance: np.float64) -> Tuple:
        """
        Saltelli estimators of the total mean and variance
        """

        variable_main_variance = sum(f_main_sample * f_new_fixing_sample) / (num_monte_carlo_points - 1) - total_mean**2
        variable_total_variance = total_variance \
                                  - sum(f_fixing_sample * f_new_fixing_sample) / (num_monte_carlo_points - 1) \
                                  + total_mean**2
        return variable_main_variance, variable_total_variance

    def compute_statistics(self, sample: np.ndarray) -> Tuple:
        """
        Computes mean and variance of a sample

        :param sample: A sample to compute statistics for.
        :return: A tuple (mean, variance).
        """
        return sample.mean(), sample.var()

    def compute_effects(self,
                        main_sample: np.ndarray=None, fixing_sample: np.ndarray=None,
                        num_monte_carlo_points: int=int(1e5)) -> Tuple:
        """
        Computes the main and total effects using Monte Carlo and a give number of samples.
        - Main effects: contribution of x_j alone to the variance of f.
        - Total effects: contribution to all Sobol terms in which x_j is involved to the variance of f.

        The (unbiased) Monte Carlo estimates are computed using:

        "A. Saltelli, Making best use of model evaluations to compute sensitivity indices, Computer Physics Com.
        608 munications, 145 (2002), pp. 280-297"

        :param main_sample: original sample that is used in the Monte Carlo computations.
        :param fixing_sample: supplementary sample that is used in the Monte Carlo computations.
        :param num_monte_carlo_points: number of points used to compute the effects.

        :return: A tuple (main effects, total effects, total variance).
        """
        if main_sample is None or fixing_sample is None:
            self.num_monte_carlo_points = num_monte_carlo_points
            self._generate_samples(self.num_monte_carlo_points)
        else:
            self.main_sample = main_sample
            self.fixing_sample = fixing_sample
            self.num_monte_carlo_points = self.main_sample.shape[0]

        f_main_sample = self.objective.f(self.main_sample)
        f_fixing_sample = self.objective.f(self.fixing_sample)

        total_mean, total_variance = self.compute_statistics(f_main_sample)
        variable_names = self.input_domain.parameter_names

        main_effects = {}
        total_effects = {}
        var_index = 0

        for variable in variable_names:
            # --- All columns are the same but the one of interest that is replaced by the original sample
            self.new_fixing_sample = self.fixing_sample.copy()
            self.new_fixing_sample[:, var_index] = self.main_sample[:, var_index]

            # --- Evaluate the objective at the new fixing sample
            f_new_fixing_sample = self.objective.f(self.new_fixing_sample)

            # --- Compute the main and total variances
            variable_main_variance, variable_total_variance = \
                self.saltelli_estimators(
                    f_main_sample, f_fixing_sample,
                    f_new_fixing_sample, self.num_monte_carlo_points,
                    total_mean, total_variance)

            # --- Compute the effects
            main_effects[variable] = variable_main_variance / total_variance
            total_effects[variable] = variable_total_variance / total_variance

            var_index += 1
        return main_effects, total_effects, total_variance


class MonteCarloSensitivity(ModelFreeMonteCarloSensitivity):
    """
    Class to compute the sensitivity coefficients of given model. This class wraps the model and calls the mean
    predictions that are used to compute the sensitivity inputs using Monte Carlo.
    """

    def __init__(self, model: IModel, input_domain: ParameterSpace)-> None:
        """
        :param model: model wrapper with the interface IModel.
        :param input_domain: space class.
        """

        self.model = model
        self.model_objective = lambda x: self.model.predict(x)[0]

        super().__init__(self.model_objective, input_domain)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .bayesian_optimization_loop import BayesianOptimizationLoop  # noqa: F401
from .unknown_constraint_bayesian_optimization_loop import UnknownConstraintBayesianOptimizationLoop  # noqa: F401



import argparse
import json
import os
import pickle
from copy import deepcopy
import numpy as np
import tarfile
import functools

from urllib.request import urlretrieve

try:
    import torch
    import torch.nn as nn
except ImportError:
    raise ImportError('pytorch is not installed. Please install it by running pip install torch torchvision')

try:
    from pybnn.bohamiann import Bohamiann
    from pybnn.util.layers import AppendLayer
except ImportError:
    raise ImportError('pybnn is not installed. Please install it by running pip install pybnn')

try:
    import GPy
    from GPy.models import BayesianGPLVM
except ImportError:
    raise ImportError('GPy is not installed. Please install it by running pip install GPy')


from emukit.examples.profet.meta_benchmarks.architecture import get_default_architecture
from emukit.examples.profet.meta_benchmarks.meta_forrester import get_architecture_forrester


def download_data(path, source='http://www.ml4aad.org/wp-content/uploads/2019/05/profet_data.tar.gz'):

    l = urlretrieve(source)[0]

    tar = tarfile.open(l)
    tar.extractall(path=path)
    tar.close()


def load_data(path, filename):
    res = json.load(open(os.path.join(path, filename), "r"))
    return np.array(res["X"]), np.array(res["Y"]), np.array(res["C"])


def normalize_Y(Y, indexD):
    max_idx = np.max(indexD)
    Y_mean = np.zeros(max_idx + 1)
    Y_std = np.zeros(max_idx + 1)
    for i in range(max_idx + 1):
        Y_mean[i] = Y[indexD == i].mean()
        Y_std[i] = Y[indexD == i].std() + 1e-8
        Y[indexD == i] = (Y[indexD == i] - Y_mean[i]) / Y_std[i]
    return Y, Y_mean[:, None], Y_std[:, None]


hidden_space = {
    'forrester': 2,
    'fcnet': 5,
    'svm': 5,
    'xgboost': 5,
    }

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--n_samples', default=1000, type=int, nargs='?',
                        help='number of samples to draw from the meta-model')
    parser.add_argument('--output_path', default="./", type=str, nargs='?',
                        help='specifies the path where the samples will be saved')
    parser.add_argument('--benchmark', default="forrester", type=str, nargs='?',
                        help='specifies the benchmark')
    parser.add_argument('--input_path', default="./", type=str, nargs='?',
                        help='path to the input data')
    parser.add_argument('--n_hidden', default=5, type=int, nargs='?',
                        help='number of hidden dimensions')
    parser.add_argument('--burnin', default=50000, type=int, nargs='?',
                        help='number of burnin steps for SGHMC')
    parser.add_argument('--download', action='store_true')
    args = parser.parse_args()

    n_samples = args.n_samples
    n_samples_task = args.n_hidden
    n_inducing_lvm = 50
    Q_h = hidden_space[args.benchmark]  # the dimensionality of the latent space
    mcmc_thining = 100
    num_burnin_steps = args.burnin
    num_steps = 100 * n_samples + 1
    lr = 1e-2
    batch_size = 5

    # load and normalize data
    if args.benchmark == "forrester":
        fname = "data_sobol_forrester.json"
        get_architecture = get_architecture_forrester
    elif args.benchmark == "svm":
        get_architecture = functools.partial(get_default_architecture, classifcation=True)
        fname = "data_sobol_svm.json"
    elif args.benchmark == "fcnet":
        get_architecture = functools.partial(get_default_architecture, classifcation=True)
        fname = "data_sobol_fcnet.json"
    elif args.benchmark == "xgboost":
        fname = "data_sobol_xgboost.json"
        get_architecture = get_default_architecture

    if args.download:
        download_data(args.input_path)

    X, Y, C = load_data(args.input_path, fname)
    if len(X.shape) == 1:
        X = X[:, None]

    n_tasks = Y.shape[0]
    n_configs = X.shape[0]
    index_task = np.repeat(np.arange(n_tasks), n_configs)
    Y_norm, _, _ = normalize_Y(deepcopy(Y.flatten()), index_task)

    # train the probabilistic encoder
    kern = GPy.kern.Matern52(Q_h, ARD=True)

    m_lvm = BayesianGPLVM(Y_norm.reshape(n_tasks, n_configs), Q_h, kernel=kern,
                          num_inducing=n_inducing_lvm)
    m_lvm.optimize(max_iters=10000, messages=1)

    ls = np.array([m_lvm.kern.lengthscale[i] for i in range(m_lvm.kern.lengthscale.shape[0])])

    # generate data to train the multi-task model
    task_features_mean = np.array(m_lvm.X.mean / ls)
    task_features_std = np.array(np.sqrt(m_lvm.X.variance) / ls)
    X_train = []
    Y_train = []
    C_train = []

    for i, xi in enumerate(X):
        for idx in range(n_tasks):
            for _ in range(n_samples_task):
                ht = task_features_mean[idx] + task_features_std[idx] * np.random.randn(ls.shape[0])

                x = np.concatenate((xi, ht), axis=0)
                X_train.append(x)
                Y_train.append(Y[idx, i])
                C_train.append(C[idx, i])

    X_train = np.array(X_train)
    Y_train = np.array(Y_train)
    C_train = np.array(C_train)
    if args.benchmark != "forrester":
        C_train = np.log(C_train)

    if args.benchmark == "xgboost":
        Y_train = np.log(Y_train)

    normalize_targets = True
    if args.benchmark == "fcnet" or args.benchmark == "svm":
        normalize_targets = False

    model_objective = Bohamiann(get_network=get_architecture, print_every_n_steps=10000,
                                normalize_output=normalize_targets)
    model_objective.train(X_train, Y_train, num_steps=num_steps + num_burnin_steps,
                          num_burn_in_steps=num_burnin_steps, keep_every=mcmc_thining,
                          lr=lr, verbose=True, batch_size=batch_size)

    if args.benchmark != "forrester":
        model_cost = Bohamiann(get_network=get_default_architecture, print_every_n_steps=10000)
        model_cost.train(X_train, C_train, num_steps=num_steps + num_burnin_steps,
                         num_burn_in_steps=num_burnin_steps, keep_every=mcmc_thining,
                         lr=lr, verbose=True, batch_size=batch_size)

    # generate samples
    sampled_h = np.zeros([n_samples, ls.shape[0]])

    os.makedirs(args.output_path, exist_ok=True)

    for i in range(n_samples):
        print("Generate sample %d" % i)
        idx = np.random.randint(n_tasks)
        ht = task_features_mean[idx] + task_features_std[idx] * np.random.randn(ls.shape[0])
        sampled_h[i] = ht
        net = model_objective.get_network(X_train.shape[1])
        net.float()

        with torch.no_grad():
            weights = model_objective.sampled_weights[i]
            for parameter, sample in zip(net.parameters(), weights):
                parameter.copy_(torch.from_numpy(sample))

        data = dict()

        data["state_dict"] = net.state_dict()
        data["x_mean"] = model_objective.x_mean
        data["x_std"] = model_objective.x_std
        if args.benchmark == "forrester" or args.benchmark == "xgboost":
            data["y_mean"] = model_objective.y_mean
            data["y_std"] = model_objective.y_std

        data["task_feature"] = ht
        pickle.dump(data, open(os.path.join(args.output_path, "sample_objective_%d.pkl" % i), "wb"))

        if args.benchmark != "forrester":
            net = model_cost.get_network(X_train.shape[1])
            net.float()

            with torch.no_grad():
                weights = model_cost.sampled_weights[i]
                for parameter, sample in zip(net.parameters(), weights):
                    parameter.copy_(torch.from_numpy(sample))

            data = dict()

            data["state_dict"] = net.state_dict()
            data["x_mean"] = model_cost.x_mean
            data["x_std"] = model_cost.x_std
            data["y_mean"] = model_cost.y_mean
            data["y_std"] = model_cost.y_std

            data["task_feature"] = ht
            pickle.dump(data, open(os.path.join(args.output_path, "sample_cost_%d.pkl" % i), "wb"))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import abc
from typing import Callable

from . import LoopState
from ..interfaces import IModel


import logging
_log = logging.getLogger(__name__)


class ModelUpdater(abc.ABC):
    @abc.abstractmethod
    def update(self, loop_state: LoopState) -> None:
        """
        Updates the training data of the model. Chooses whether to update hyper-parameters and how to do it

        :param loop_state: Object that contains current state of the loop
        """
        pass


class NoopModelUpdater(ModelUpdater):
    def update(self, loop_state: LoopState) -> None:
        """
        Dummy model updater that does nothing. It can be used for example for random search

        :param loop_state: Object that contains current state of the loop
        """
        pass


class FixedIntervalUpdater(ModelUpdater):
    """ Updates hyper-parameters every nth iteration, where n is defined by the user """
    def __init__(self, model: IModel, interval: int=1, targets_extractor_fcn: Callable=None) -> None:
        """
        :param model: Emukit emulator model
        :param interval: Number of function evaluations between optimizing model hyper-parameters
        :param targets_extractor_fcn: A function that takes in loop state and returns the training targets.
                                      Defaults to a function returning loop_state.Y
        """
        self.model = model
        self.interval = interval

        if targets_extractor_fcn is None:
            self.targets_extractor_fcn = lambda loop_state: loop_state.Y
        else:
            self.targets_extractor_fcn = targets_extractor_fcn

    def update(self, loop_state: LoopState) -> None:
        """
        :param loop_state: Object that contains current state of the loop
        """
        targets = self.targets_extractor_fcn(loop_state)
        self.model.set_data(loop_state.X, targets)
        if (loop_state.iteration % self.interval) == 0:
            _log.info("Updating parameters of the model")
            self.model.optimize()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .acquisition_optimizer import AcquisitionOptimizerBase  # noqa: F401
from .context_manager import ContextManager  # noqa: F401
from .gradient_acquisition_optimizer import GradientAcquisitionOptimizer  # noqa: F401
from .local_search_acquisition_optimizer import LocalSearchAcquisitionOptimizer  # noqa: F401
from .multi_source_acquisition_optimizer import MultiSourceAcquisitionOptimizer  # noqa: F401
from .random_search_acquisition_optimizer import RandomSearchAcquisitionOptimizer  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .quadrature_kernels import QuadratureKernel  # noqa: F401
from .bounds import BoxBounds  # noqa: F401
from .quadrature_rbf import QuadratureRBFLebesgueMeasure, QuadratureRBFIsoGaussMeasure, QuadratureRBFUniformMeasure  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .squared_correlation import SquaredCorrelation  # noqa: F401
from .squared_correlation import SquaredCorrelation as IntegralVarianceReduction  # noqa: F401
from .mutual_information import MutualInformation  # noqa: F401
from .uncertainty_sampling import UncertaintySampling  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .linear_model import GPyLinearMultiFidelityModel  # noqa: F401
from .non_linear_multi_fidelity_model import NonLinearMultiFidelityModel  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np


class UserFunctionResult(object):
    """
    A class that records the inputs, outputs and meta-data of an evaluation of the user function.
    """
    def __init__(self, X: np.ndarray, Y: np.ndarray, **kwargs) -> None:
        """
        :param X: Function input. Shape: (function input dimension,)
        :param Y: Function output(s). Shape: (function output dimension,)
        :param kwargs: Extra outputs of the UserFunction to store. Shape: (extra output dimension,)
        """
        if X.ndim != 1:
            raise ValueError("x is expected to be 1-dimensional, actual dimensionality is {}".format(X.ndim))

        if Y.ndim != 1:
            raise ValueError("y is expected to be 1-dimensional, actual dimensionality is {}".format(Y.ndim))

        self.extra_outputs = dict()
        for (key, val) in kwargs.items():
            if val.ndim != 1:
                raise ValueError('Key word arguments must be 1-dimensional but {} is {}d'.format(key, val.ndim))
            self.extra_outputs[key] = val

        self.X = X
        self.Y = Y

    def __getattr__(self, item):
        """
        Allow extra output values to be accessed as an attribute

        :param item: The name of the extra output to be accessed
        :return: The value of the extra output
        """
        return self.extra_outputs[item]

    def __repr__(self):
        return "UserFunctionResult(X: {}, Y: {}, extra_outputs: {})".format(self.X, self.Y, self.extra_outputs)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from .latin_design import LatinDesign  # noqa: F401
from .random_design import RandomDesign  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np

from emukit.core import ParameterSpace, ContinuousParameter


def branin_function():
    """
    Two-dimensional Branin, often used as an optimization benchmark.

    Based on: https://www.sfu.ca/~ssurjano/branin.html

    .. math::
        f(\mathbf{x}) = (x_2 - b x_1 ^ 2 + c x_1 - r) ^ 2 + s(1 - t) \cos(x_1) + s

    where:

    .. math::
        b = 5.1 / (4 \pi ^ 2)

        c = 5 /\pi

        r = 6

        s = 10

        t = 1 / (8\pi)
    """

    parameter_space = ParameterSpace([ContinuousParameter('x1', -5, 10), ContinuousParameter('x2', 0, 15)])
    return _branin, parameter_space


def _branin(x):
    """
    :param x: n_points x 2 array of input locations to evaluate
    :return: n_points x 1 array of function evaluations
    """
    a = 1
    b = 5.1 / (4 * np.pi**2)
    c = 5 / np.pi
    r = 6
    s = 10
    t = 1 / (8 * np.pi)
    x1 = x[:, 0]
    x2 = x[:, 1]
    y = a * (x2 - b * x1**2 + c * x1 - r)**2 + s * (1 - t) * np.cos(x1) + s
    return y[:, None]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from .acquisitions import ModelVariance
from ..core.acquisition import Acquisition
from ..core.interfaces.models import IModel
from ..core.loop import OuterLoop, SequentialPointCalculator, FixedIntervalUpdater
from ..core.loop.candidate_point_calculators import GreedyBatchPointCalculator
from ..core.loop.loop_state import create_loop_state
from ..core.optimization import AcquisitionOptimizerBase
from ..core.optimization import GradientAcquisitionOptimizer
from ..core.parameter_space import ParameterSpace


class ExperimentalDesignLoop(OuterLoop):
    def __init__(self, space: ParameterSpace, model: IModel, acquisition: Acquisition = None, update_interval: int = 1,
                 batch_size: int = 1, acquisition_optimizer: AcquisitionOptimizerBase = None):
        """
        An outer loop class for use with Experimental design

        :param space: Definition of domain bounds to collect points within
        :param model: The model that approximates the underlying function
        :param acquisition: experimental design acquisition function object. Default: ModelVariance acquisition
        :param update_interval: How many iterations pass before next model optimization
        :param batch_size: Number of points to collect in a batch. Defaults to one.
        :param acquisition_optimizer: Optimizer selecting next evaluation points
                                      by maximizing acquisition.
                                      Gradient based optimizer is used if None.
                                      Defaults to None.
        """

        if acquisition is None:
            acquisition = ModelVariance(model)

        # This AcquisitionOptimizer object deals with optimizing the acquisition to find the next point to collect
        if acquisition_optimizer is None:
            acquisition_optimizer = GradientAcquisitionOptimizer(space)

        # Construct emukit classes
        if batch_size == 1:
            candidate_point_calculator = SequentialPointCalculator(acquisition, acquisition_optimizer)
        elif batch_size > 1:
            candidate_point_calculator = \
                GreedyBatchPointCalculator(model, acquisition, acquisition_optimizer, batch_size)
        else:
            raise ValueError('Batch size value of ' + str(batch_size) + ' is invalid.')

        model_updater = FixedIntervalUpdater(model, update_interval)
        loop_state = create_loop_state(model.X, model.Y)

        super().__init__(candidate_point_calculator, model_updater, loop_state)

        self.model = model



import logging
import numpy as np
from functools import partial
from typing import Callable, List, Tuple, Union

from ...core import ParameterSpace
from ...core.loop import LoopState, UserFunction, UserFunctionWrapper
from ...core.initial_designs import RandomDesign
from ...core.initial_designs.base import ModelFreeDesignBase
from .benchmark_result import BenchmarkResult
from .metrics import Metric

_log = logging.getLogger(__name__)


class Benchmarker:
    def __init__(self, loops_with_names: List[Tuple[str, Callable]], test_function: Union[Callable, UserFunction],
                 parameter_space: ParameterSpace, metrics: List[Metric], initial_design: ModelFreeDesignBase=None):
        """
        :param loops_with_names: A list of tuples where the first entry is the name of the loop and the second is a
                                 function that takes in initial x and y training data and returns a loop to be
                                 benchmarked
        :param test_function: The function to benchmark the loop against
        :param parameter_space: Parameter space describing the input domain of the function to be benchmarked against
        :param metrics: List of metric objects that assess the performance of the loop at every iteration
        :param initial_design: An object that returns a set of samples in the input domain that are used as the initial
                               data set
        """

        self.loop_names = [loop[0] for loop in loops_with_names]
        self.loops = [loop[1] for loop in loops_with_names]

        if isinstance(test_function, UserFunction):
            self.test_function = test_function
        else:
            self.test_function = UserFunctionWrapper(test_function)
        self.parameter_space = parameter_space

        if initial_design is None:
            initial_design = RandomDesign(parameter_space)
        self.initial_design = initial_design
        self.metrics = metrics
        self.metric_names = [metric.name for metric in metrics]

        if len(set(self.metric_names)) != len(self.metric_names):
            raise ValueError('Names of metrics are not unique')

    def run_benchmark(self, n_initial_data: int=10, n_iterations: int=10, n_repeats: int=10) -> BenchmarkResult:
        """
        Runs the benchmarking. For each initial data set, every loop is created and run for the specified number of
        iterations and the results are collected.

        :param n_initial_data: Number of points in the initial data set
        :param n_iterations: Number of iterations to run the loop for
        :param n_repeats: Number of times to run each loop with a different initial data set
        :return: An instance of BenchmarkResult that contains all the tracked metrics for each loop
        """
        result = BenchmarkResult(self.loop_names, n_repeats, self.metric_names)
        for j in range(n_repeats):
            initial_loop_state = self._create_initial_loop_state(n_initial_data)
            for i, (loop, loop_name) in enumerate(zip(self.loops, self.loop_names)):
                _log.info('Benchmarking loop ' + str(i) + ' for repeat ' + str(j))

                this_loop = loop(initial_loop_state)
                this_loop.loop_state.metrics = dict()
                self._subscribe_metrics_to_loop_events(this_loop)

                this_loop.run_loop(self.test_function, n_iterations)

                for metric_name, metric_value in this_loop.loop_state.metrics.items():
                    result.add_results(loop_name, j, metric_name, metric_value)
        return result

    def _subscribe_metrics_to_loop_events(self, outer_loop):
        """
        Subscribe metric calls to events on outer loop object
        """
        if self.metrics is not None:
            for metric in self.metrics:
                metric.reset()

                def update_metric(loop, loop_state, metric):
                    value = metric.evaluate(loop, loop_state)
                    _add_value_to_metrics_dict(loop_state, value, metric.name)

                # Subscribe to events
                func = partial(update_metric, metric=metric)
                outer_loop.loop_start_event.append(func)
                outer_loop.iteration_end_event.append(func)

    def _create_initial_loop_state(self, n_initial_data):
        x_init = self.initial_design.get_samples(n_initial_data)
        results = self.test_function.evaluate(x_init)
        return LoopState(results)


def _add_value_to_metrics_dict(loop_state, value, key_name):
    """
    Add new metric evaluation to dictionary in loop state using the metric name as the key in the dictionary
    """

    if key_name in loop_state.metrics:
        # Array already exists - append new value
        loop_state.metrics[key_name] = np.concatenate([loop_state.metrics[key_name], [value]], axis=0)
    else:
        # Initialise array
        loop_state.metrics[key_name] = np.array([value])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Callable, List, Tuple

import numpy as np
import scipy.optimize

from .. import ParameterSpace
from ..constraints import IConstraint, LinearInequalityConstraint, NonlinearInequalityConstraint
from .context_manager import ContextManager


class Optimizer(object):
    """
    Class for a general acquisition optimizer.
    """

    def __init__(self, bounds: List[Tuple]):
        """
        :param bounds: List of min/max values for each dimension of x
        """
        self.bounds = bounds

    def optimize(self, x0: np.ndarray, f: Callable=None, df: Callable=None, f_df: Callable=None) \
            -> Tuple[np.ndarray, np.ndarray]:
        """
        :param x0: initial point for a local optimizer.
        :param f: function to optimize.
        :param df: gradient of the function to optimize.
        :param f_df: returns both the function to optimize and its gradient.
        :return: Location of optimum and value at optimum
        """
        raise NotImplementedError("The optimize method is not implemented in the parent class.")


class OptLbfgs(Optimizer):
    """
    Wrapper for l-bfgs-b to use the true or the approximate gradients.
    """

    def __init__(self, bounds, max_iterations=1000):
        super(OptLbfgs, self).__init__(bounds)
        self.max_iterations = max_iterations

    def optimize(self, x0: np.ndarray, f: Callable=None, df: Callable=None, f_df: Callable=None) \
            -> Tuple[np.ndarray, np.ndarray]:
        """
        :param x0: initial point for a local optimizer.
        :param f: function to optimize.
        :param df: gradient of the function to optimize.
        :param f_df: returns both the function to optimize and its gradient.
        :return: Location of optimum and value at optimum
        """

        if f_df is None and df is not None: f_df = lambda x: (float(f(x)), df(x))
        if f_df is not None:
            def _f_df(x):
                return f(x), f_df(x)[1][0]

        if f_df is None and df is None:
            res = scipy.optimize.fmin_l_bfgs_b(f, x0=x0, bounds=self.bounds, approx_grad=True, maxiter=self.max_iterations)
        else:
            res = scipy.optimize.fmin_l_bfgs_b(_f_df, x0=x0, bounds=self.bounds, maxiter=self.max_iterations)

        # We check here if the the optimizer moved. It it didn't we report x0 and f(x0) as scipy can return NaNs
        if res[2]['task'] == b'ABNORMAL_TERMINATION_IN_LNSRCH':
            result_x = np.atleast_2d(x0)
            result_fx = np.atleast_2d(f(x0))
        else:
            result_x = np.atleast_2d(res[0])
            result_fx = np.atleast_2d(res[1])

        return result_x, result_fx


def apply_optimizer(optimizer: Optimizer, x0: np.ndarray, space: ParameterSpace, f: Callable=None, df: Callable=None,
                    f_df: Callable=None, context_manager: ContextManager=None) -> Tuple[np.ndarray, np.ndarray]:
    """
    Optimizes f using the optimizer supplied, deals with potential context variables.

    :param optimizer: The optimizer object that will perform the optimization
    :param x0: initial point for a local optimizer (x0 can be defined with or without the context included).
    :param f: function to optimize.
    :param df: gradient of the function to optimize.
    :param f_df: returns both the function to optimize and its gradient.
    :param context_manager: If provided, x0 (and the optimizer) operates in the space without the context
    :param space: Parameter space describing input domain, including any context variables
    :return: Location of optimum and value at optimum
    """

    if context_manager is None:
        context_manager = ContextManager(space, {})

    # Compute new objective that inputs non context variables but takes into account the values of the context ones.
    # It does nothing if no context is passed
    problem = OptimizationWithContext(x0=x0, f=f, df=df, f_df=f_df, context_manager=context_manager)

    add_context = lambda x: context_manager.expand_vector(x)

    # Optimize point
    if f is None:
        f_no_context = None
    else:
        f_no_context = problem.f_no_context

    if df is None:
        df_no_context = None
    else:
        df_no_context = problem.df_no_context

    if f_df is None:
        f_df_no_context = None
    else:
        f_df_no_context = problem.f_df_no_context

    optimized_x, _ = optimizer.optimize(problem.x0_no_context, f_no_context, df_no_context, f_df_no_context)

    # Add context and round according to the type of variables of the design space
    suggested_x_with_context = add_context(optimized_x)
    suggested_x_with_context_rounded = space.round(suggested_x_with_context)

    if f is None:
        f_opt, _ = f_df(suggested_x_with_context_rounded)
    else:
        f_opt = f(suggested_x_with_context_rounded)
    return suggested_x_with_context_rounded, f_opt


class OptimizationWithContext(object):

    def __init__(self, x0: np.ndarray, f: Callable, df: Callable=None, f_df: Callable=None,
                 context_manager: ContextManager = None):
        """
        Constructor of an objective function that takes as input a vector x of the non context variables
        and returns a value in which the context variables have been fixed.
        """
        self.x0 = np.atleast_2d(x0)
        self.f = f
        self.df = df
        self.f_df = f_df
        self.context_manager = context_manager

        if not context_manager:
            self.x0_no_context = x0
            self.f_no_context = self.f
            self.df_no_context = self.df
            self.f_df_no_context = self.f_df
        else:
            self.x0_no_context = self.x0[:, self.context_manager.non_context_idxs]
            self.f_no_context = self.f_no_context
            if self.f_df is None:
                self.df_no_context = None
                self.f_df_no_context = None
            else:
                self.df_no_context = self.df_no_context
                self.f_df_no_context = self.f_df_no_context

    def f_no_context(self, x: np.ndarray) -> np.ndarray:
        """
        Wrapper of optimization objective function which deals with adding context variables to x

        :param x: Input without context variables
        """
        x = np.atleast_2d(x)
        xx = self.context_manager.expand_vector(x)
        if x.shape[0] == 1:
            return self.f(xx)[0]
        else:
            return self.f(xx)

    def df_no_context(self, x: np.ndarray) -> np.ndarray:
        """
        Wrapper of the derivative of optimization objective function which deals with adding context variables to x

        :param x: Input without context variables
        """
        x = np.atleast_2d(x)
        xx = self.context_manager.expand_vector(x)
        _, df_no_context_xx = self.f_df(xx)
        df_no_context_xx = df_no_context_xx[:, np.array(self.context_manager.non_context_idxs)]
        return df_no_context_xx

    def f_df_no_context(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Wrapper of optimization objective function and its derivative which deals with adding context variables to x

        :param x: Input without context variables
        """
        x = np.atleast_2d(x)
        xx = self.context_manager.expand_vector(x)
        f_no_context_xx, df_no_context_xx = self.f_df(xx)
        df_no_context_xx = df_no_context_xx[:, np.array(self.context_manager.non_context_idxs)]
        return f_no_context_xx, df_no_context_xx


class OptTrustRegionConstrained(Optimizer):
    """
    Wrapper for Trust-Region Constrained algorithm that can deal with non-linear constraints
    """

    def __init__(self, bounds: List[Tuple], constraints: List[IConstraint], max_iterations: int=1000):
        """
        :param bounds: List of tuples containing (lower_bound, upper_bound) for each parameter
        :param constraints: List of constraints, can contain a mix of linear and non-linear constraints
        :param max_iterations: Maximum number of iterations before the optimizer is stopped
        """
        super().__init__(bounds)
        self.max_iterations = max_iterations
        self.constraints = _get_scipy_constraints(constraints)

    def optimize(self, x0: np.ndarray, f: Callable=None, df: Callable=None, f_df: Callable=None) \
            -> Tuple[np.ndarray, np.ndarray]:
        """
        Run Trust region constrained optimization algorithm

        :param x0: Initial start point
        :param f: Function to optimize
        :param df: Derivative of function to optimize
        :param f_df: Function returning both value of objective and its gradient
        :return: Location of optimum and function value at optimum
        """

        if (f is None) and (f_df is None):
            raise ValueError('Neither f nor f_df are supplied - you must supply an objective function')

        # If f not supplied, make lambda that returns objective only from f_df
        if f is None:
            f = lambda x: f_df(x)[0]

        if df is None and f_df is not None:
            # If df not supplied and f_df is, make lambda that returns gradient only from f_df
            df_1d = lambda x: f_df(x)[1][0, :]
        elif df is not None:
            # If df is supplied, convert the 2d output to 1d
            df_1d = lambda x: df(x)[0, :]
        else:
            # Gradient not supplied
            df_1d = None

        options = {'maxiter': self.max_iterations}

        if df_1d is None:
            res = scipy.optimize.minimize(f, x0=x0[0, :], method='trust-constr', bounds=self.bounds, jac='2-point',
                                          options=options, constraints=self.constraints, hess=scipy.optimize.BFGS())
        else:
            res = scipy.optimize.minimize(f, x0=x0[0, :], method='trust-constr', bounds=self.bounds, jac=df_1d,
                                          options=options, constraints=self.constraints, hess=scipy.optimize.BFGS())

        result_x = np.atleast_2d(res.x)
        result_fx = np.atleast_2d(res.fun)
        return result_x, result_fx


def _get_scipy_constraints(constraint_list: List[IConstraint]) -> List:
    """
    Converts list of emukit constraint objects to list of scipy constraint objects

    :param constraint_list: List of Emukit constraint objects
    :return: List of scipy constraint objects
    """

    scipy_constraints = []
    for constraint in constraint_list:
        if isinstance(constraint, NonlinearInequalityConstraint):
            if constraint.jacobian_fun is None:
                # No jacobian supplied -> tell scipy to use finite difference method
                jacobian = '2-point'
            else:
                # Jacobian is supplied -> tell scipy to use it
                jacobian = constraint.jacobian_fun

            scipy_constraints.append(
                scipy.optimize.NonlinearConstraint(constraint.fun, constraint.lower_bound, constraint.upper_bound,
                                                   jacobian))
        elif isinstance(constraint, LinearInequalityConstraint):
            scipy_constraints.append(scipy.optimize.LinearConstraint(constraint.constraint_matrix,
                                                                     constraint.lower_bound,
                                                                     constraint.upper_bound))
        else:
            raise ValueError('Constraint type {} not recognised'.format(type(constraint)))
    return scipy_constraints



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import itertools
from typing import List, Optional, Tuple

import numpy as np

from .constraints import IConstraint
from .discrete_parameter import InformationSourceParameter
from .parameter import Parameter


class ParameterSpace(object):
    """
    Represents parameter space for a given problem.
    """

    def __init__(self, parameters: List, constraints: Optional[List[IConstraint]]=None):
        """
        Creates a new instance of a parameter space.

        :param parameters: A list of parameters in the space.
        :param constraints: A list of constraints on the input domain
        """
        self._parameters = parameters

        if constraints:
            self.constraints = constraints
        else:
            self.constraints = []

        # Check no more than one InformationSource parameter
        source_parameter = [param for param in self.parameters if isinstance(param, InformationSourceParameter)]
        if len(source_parameter) > 1:
            raise ValueError('More than one source parameter found')

        # Check uniqueness of parameter names
        names = self.parameter_names
        if not len(names) == len(set(names)):
            raise ValueError('Parameter names are not unique')

    def find_parameter_index_in_model(self, parameter_name: str) -> List[int]:
        """
        Find the indices of the encoding of the specified parameter in the input vector

        :param parameter_name: Parameter name to find indices for
        :return: List of indices
        """
        i_start = 0
        for param in self._parameters:
            if param.name == parameter_name:
                return list(range(i_start, i_start + param.dimension))
            else:
                i_start += param.dimension
        raise ValueError('Parameter {} not found'.format(parameter_name))

    @property
    def dimensionality(self):
        return sum([p.dimension for p in self._parameters])

    @property
    def parameters(self) -> List:
        """
        Returns the list of parameters in the space.
        """

        return self._parameters

    @property
    def parameter_names(self) -> List:
        """
        Returns the list of names of parameters in the space.
        """

        return [p.name for p in self._parameters]

    def get_parameter_by_name(self, name: str) -> Parameter:
        """
        Returns parameter with the given name

        :param name: Parameter name
        :returns: A parameter object
        """

        for param in self.parameters:
            if param.name == name:
                return param
        raise ValueError('Parameter with name ' + name + ' not found.')

    def get_bounds(self) -> List[Tuple]:
        """
        Returns a list of tuples containing the min and max value each parameter can take.

        If the parameter space contains categorical variables, the min and max values correspond to each variable used
        to encode the categorical variables.
        """

        # bounds is a list of lists
        bounds = [param.bounds for param in self.parameters]
        # Convert list of lists to one list
        return list(itertools.chain.from_iterable(bounds))

    def round(self, x: np.ndarray) -> np.ndarray:
        """
        Rounds given values x to closest valid values within the space.

        :param x: A 2d array of values to be rounded
        :returns: A 2d array of rounded values
        """

        x_rounded = []
        current_idx = 0
        for param in self.parameters:
            param_columns = x[:, current_idx:(current_idx + param.dimension)]
            x_rounded.append(param.round(param_columns))
            current_idx += param.dimension

        return np.column_stack(x_rounded)

    def check_points_in_domain(self, x: np.ndarray) -> np.ndarray:
        """
        Checks that each column of x lies in the domain of the corresponding parameter

        :param x: 2d numpy array of points to check
        :return: A 1d numpy array which contains a boolean indicating whether each point is in domain
        """
        len_encoding = sum(len(param.model_parameters) for param in self.parameters)
        if x.shape[1] != len_encoding:
            raise ValueError('x should have number of columns equal to the sum'
                             'of all parameter encodings, expected {} actual {}'
                             .format(x.shape[1], len_encoding))

        in_domain = np.ones(x.shape[0], dtype=bool)
        encoding_index = 0
        for param in self._parameters:
            # First check if this particular parameter is in domain
            param_in_domain = [
                param.check_in_domain(x[[point_ix], encoding_index:(encoding_index + param.dimension)])
                for point_ix in range(x.shape[0])]
            # Set in_domain to be False if this parameter or any previous parameter is out of domain
            in_domain = np.all([in_domain, param_in_domain], axis=0)
            encoding_index += param.dimension
        return in_domain

    def sample_uniform(self, point_count: int) -> np.ndarray:
        """
        Generates multiple uniformly distributed random parameter points.

        :param point_count: number of data points to generate.
        :returns: Generated points with shape (point_count, num_features)
        """
        parameter_samples = [param.sample_uniform(point_count) for param in self.parameters]
        return np.hstack(parameter_samples)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from ...core.loop.outer_loop import OuterLoop
from ...core.loop.candidate_point_calculators import RandomSampling
from ...core.loop.loop_state import create_loop_state
from ...core.parameter_space import ParameterSpace
from ...core.loop.model_updaters import NoopModelUpdater


class RandomSearch(OuterLoop):
    def __init__(self, space: ParameterSpace, x_init: np.ndarray = None,
                 y_init: np.ndarray = None, cost_init: np.ndarray = None):

        """
        Simple loop to perform random search where in each iteration points are sampled uniformly at random
        over the input space.

        :param space: Input space where the optimization is carried out.
        :param x_init: 2d numpy array of shape (no. points x no. input features) of initial X data
        :param y_init: 2d numpy array of shape (no. points x no. targets) of initial Y data
        :param cost_init: 2d numpy array of shape (no. points x no. targets) of initial cost of each function evaluation
        """

        model_updaters = NoopModelUpdater()

        candidate_point_calculator = RandomSampling(parameter_space=space)

        if x_init is not None and y_init is not None:
            loop_state = create_loop_state(x_init, y_init, cost=cost_init)
        else:
            loop_state = None

        super().__init__(candidate_point_calculator, model_updaters, loop_state=loop_state)



class EventHandler(list):
    """
    A list of callable objects. Calling an instance of this will cause a call to each item in the list in ascending
    order by index.

    Code taken from: https://stackoverflow.com/a/2022629

    To subscribe to the event simply append a function to the event handler:
    ``event_handler.append(fcn_to_call_on_event)``
    """
    def __call__(self, *args, **kwargs):
        for f in self:
            f(*args, **kwargs)

    def __repr__(self):
        return "Event(%s)" % list.__repr__(self)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import Tuple, List

from ...core.optimization.context_manager import ContextManager


class IntegrationMeasure:
    """An abstract class for a probability measure with a density"""

    def __init__(self, name: str):
        """
        :param name: Name of the integration measure
        """
        self.name = name

    def compute_density(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the density at point x
        :param x: points at which density is computed, shape (num_points, num_dimensions)
        :return: the density at x, shape (num_points, )
        """
        raise NotImplementedError

    def compute_density_gradient(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the gradient of the density at point x.
        Might be needed for some acquisition functions.

        :param x: points at which the gradient is computed, shape (num_points, num_dimensions)
        :return: the gradient of the density at x, shape (num_points, num_dimensions)
        """
        raise NotImplementedError

    def get_box(self) -> List[Tuple[float, float]]:
        """
        Meaningful box-bounds around the measure. Outside this box, the measure should be virtually zero.

        :return: box in which the measure lies. List of D tuples, where D is the dimensionality and the tuples contain
        the lower and upper bounds of the box i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]
        """
        raise NotImplementedError

    @property
    def can_sample(self) -> bool:
        """
        Indicates whether probability measure has sampling available.
        :return: True if sampling is available
        """
        raise NotImplementedError

    def get_samples(self, num_samples: int, context_manager: ContextManager = None) -> np.ndarray:
        """
        Samples from the probability distribution defined by the integration measure.

        :param num_samples: number of samples
        :param context_manager: The context manager that contains variables to fix and the values to fix them to. If a
        context is given, this method samples from the conditional distribution.
        :return: samples, shape (num_samples, num_dimensions)
        """
        raise NotImplementedError


class UniformMeasure(IntegrationMeasure):
    """The Uniform measure"""

    def __init__(self, bounds: List[Tuple[float, float]]):
        """
        :param bounds: List of D tuples, where D is the dimensionality of the domain and the tuples contain the lower
        and upper bounds of the box defining the uniform measure i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]
        """
        super().__init__('UniformMeasure')

        # checks if lower bounds are smaller than upper bounds.
        for (lb_d, ub_d) in bounds:
            if lb_d >= ub_d:
                raise ValueError("Upper bound of uniform measure must be larger than lower bound. Found a pair "
                                 "containing ({}, {}).".format(lb_d, ub_d))

        self.bounds = bounds
        # uniform measure has constant density which is computed here.
        self.density = self._compute_constant_density()

    def _compute_constant_density(self) -> float:
        differences = np.array([x[1] - x[0] for x in self.bounds])
        volume = np.prod(differences)

        if volume <= 0:
            raise NumericalPrecisionError("Domain volume of uniform measure is not positive. Its value is {}.".format(
                volume))
        return float(1. / volume)

    @property
    def can_sample(self) -> bool:
        """
        Indicates whether probability measure has sampling available.
        :return: True if sampling is available
        """
        return True

    def compute_density(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the density at point x
        :param x: points at which density is computed, shape (num_points, num_dimensions)
        :return: the density at x, shape (num_points, )
        """
        # compute density: (i) check if points are inside the box. (ii) multiply this bool value with density.
        bounds_lower = np.array([b[0] for b in self.bounds])
        bounds_upper = np.array([b[1] for b in self.bounds])
        inside_lower = 1 - (x < bounds_lower)  # contains 1 if element in x is above its lower bound, 0 otherwise
        inside_upper = 1 - (x > bounds_upper)  # contains 1 if element in x is below its upper bound, 0 otherwise
        # contain True if element in x is inside box, False otherwise. This array multiplied with the constant density
        # as done in the return statement yields self.density for a point inside the box and 0 otherwise.
        inside_upper_lower = (inside_lower * inside_upper).sum(axis=1) == x.shape[1]
        return inside_upper_lower * self.density

    def compute_density_gradient(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the gradient of the density at point x
        :param x: points at which the gradient is computed, shape (num_points, num_dimensions)
        :return: the gradient of the density at x, shape (num_points, num_dimensions)
        """
        return np.zeros(x.shape)

    def get_box(self) -> List[Tuple[float, float]]:
        """
        Meaningful box-bounds around the measure. Outside this box, the measure should be virtually zero.

        :return: box in which the measure lies. List of D tuples, where D is the dimensionality and the tuples contain
        the lower and upper bounds of the box i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]
        """
        return self.bounds

    def get_samples(self, num_samples: int, context_manager: ContextManager=None) -> np.ndarray:
        """
        Samples from the uniform distribution.

        :param num_samples: number of samples
        :param context_manager: The context manager that contains variables to fix and the values to fix them to. If a
        context is given, this method samples from the conditional distribution.
        :return: samples, shape (num_samples, num_dimensions)
        """

        D = len(self.bounds)
        bounds = np.asarray(self.bounds)

        samples = np.random.rand(num_samples, D) * (bounds[:, 1] - bounds[:, 0]) + bounds[:, 0]

        if context_manager is not None:
            samples[:, context_manager.context_idxs] = context_manager.context_values

        return samples


class IsotropicGaussianMeasure(IntegrationMeasure):
    """
    The isotropic Gaussian measure.

    An isotropic Gaussian is a Gaussian with scalar co-variance matrix. The density is
    :math:`p(x)=(2\pi\sigma^2)^{-\frac{D}{2}} e^{-\frac{1}{2}\frac{\|x-\mu\|^2}{\sigma^2}}`
    """

    def __init__(self, mean: np.ndarray, variance: float):
        """
        :param mean: the mean of the Gaussian, shape (num_dimensions, )
        :param variance: the scalar variance of the isotropic covariance matrix of the Gaussian.
        """
        super().__init__('GaussianMeasure')
        # check mean
        if not isinstance(mean, np.ndarray):
            raise TypeError('Mean must be of type numpy.ndarray, {} given.'.format(type(mean)))

        if mean.ndim != 1:
            raise ValueError('Dimension of mean must be 1, dimension {} given.'.format(mean.ndim))

        # check covariance
        if not isinstance(variance, float):
            raise TypeError('Variance must be of type float, {} given.'.format(type(variance)))

        if not variance > 0:
            raise ValueError('Variance must be positive, current value is {}.'.format(variance))

        self.mean = mean
        self.variance = variance
        self.num_dimensions = mean.shape[0]

    @property
    def full_covariance_matrix(self):
        return self.variance * np.eye(self.num_dimensions)

    @property
    def can_sample(self) -> bool:
        """
        Indicates whether probability measure has sampling available.
        :return: True if sampling is available
        """
        return True

    def compute_density(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the density at point x
        :param x: points at which density is computed, shape (num_points, num_dimensions)
        :return: the density at x, shape (num_points, )
        """
        factor = (2 * np.pi * self.variance) ** (self.num_dimensions / 2)
        scaled_diff = (x - self.mean) / (np.sqrt(2 * self.variance))
        return np.exp(- np.sum(scaled_diff ** 2, axis=1)) / factor

    def compute_density_gradient(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the gradient of the density at point x
        :param x: points at which the gradient is computed, shape (num_points, num_dimensions)
        :return: the gradient of the density at x, shape (num_points, num_dimensions)
        """
        values = self.compute_density(x)
        return ((- values / self.variance) * (x - self.mean).T).T

    def get_box(self) -> List[Tuple[float, float]]:
        """
        Meaningful box-bounds around the measure. Outside this box, the measure should be virtually zero.

        :return: box in which the measure lies. List of D tuples, where D is the dimensionality and the tuples contain
        the lower and upper bounds of the box i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]
        """
        # Note: the factor 10 is somewhat arbitrary but well motivated. If this method is used to get a box for
        # data-collection, the box will be 2x 10 standard deviations wide in all directions, centered around the mean.
        # Outside the box the density is virtually zero.
        factor = 10
        lower = self.mean - factor * np.sqrt(self.variance)
        upper = self.mean + factor * np.sqrt(self.variance)
        return list(zip(lower, upper))

    def get_samples(self, num_samples: int, context_manager: ContextManager=None) -> np.ndarray:
        """
        Samples from the isotropic Gaussian distribution.

        :param num_samples: number of samples
        :param context_manager: The context manager that contains variables to fix and the values to fix them to. If a
        context is given, this method samples from the conditional distribution.
        :return: samples, shape (num_samples, num_dimensions)
        """
        samples = self.mean + np.sqrt(self.variance) * np.random.randn(num_samples, self.num_dimensions)

        if context_manager is not None:
            # since the Gaussian is isotropic, fixing the value after sampling the joint is equal to sampling the
            # conditional.
            samples[:, context_manager.context_idxs] = context_manager.context_values

        return samples


class NumericalPrecisionError(Exception):
    pass



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import numpy as np

from .. import ParameterSpace


class ModelFreeDesignBase(object):
    """
    Base class for all model free experiment designs
    """
    def __init__(self, parameter_space: ParameterSpace):
        """
        :param parameter_space: The parameter space to generate design for.
        """
        self.parameter_space = parameter_space

    def get_samples(self, point_count: int) -> np.ndarray:
        """
        Generates requested amount of points.

        :param point_count: Number of points required.
        :return: A numpy array with shape (point_count x space_dim)
        """
        raise NotImplementedError("Subclasses should implement this method.")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .hennig1D import hennig1D  # noqa: F401
from .circular_gaussian import circular_gaussian  # noqa: F401
from .sombrero2D import sombrero2D  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple, Union

import scipy.stats
import numpy as np

from ...core.interfaces import IModel, IDifferentiable
from ...core.acquisition import Acquisition


class ProbabilityOfFeasibility(Acquisition):

    def __init__(self, model: Union[IModel, IDifferentiable], jitter: float = float(0)) -> None:
        """
        This acquisition computes for a given input point the probability of satisfying the constraint
        C<0. For more information see:

        Michael A. Gelbart, Jasper Snoek, and Ryan P. Adams,
        Bayesian Optimization with Unknown Constraints,
        https://arxiv.org/pdf/1403.5607.pdf

        :param model: The underlying model that provides the predictive mean and variance for the given test points
        :param jitter: Jitter to balance exploration / exploitation
        """
        self.model = model
        self.jitter = jitter

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the probability of of satisfying the constraint C<0.
        :param x: points where the acquisition is evaluated, shape (number of points, number of dimensions).
        :return: numpy array with the probability of satisfying the constraint at the points x.
        """
        mean, variance = self.model.predict(x)
        mean += self.jitter

        standard_deviation = np.sqrt(variance)
        cdf = scipy.stats.norm.cdf(0, mean, standard_deviation)
        return cdf

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple:
        """
        Computes the  probability of of satisfying the constraint C<0.
        :param x: points where the acquisition is evaluated, shape (number of points, number of dimensions).
        :return: tuple of numpy arrays with the probability of satisfying the constraint at the points x 
        and its gradient.
        """
        mean, variance = self.model.predict(x)
        standard_deviation = np.sqrt(variance)

        dmean_dx, dvariance_dx = self.model.get_prediction_gradients(x)
        dstandard_devidation_dx = dvariance_dx / (2 * standard_deviation)

        mean += self.jitter
        u = - mean / standard_deviation
        pdf = scipy.stats.norm.pdf(0, mean, standard_deviation)
        cdf = scipy.stats.norm.cdf(0, mean, standard_deviation)
        dcdf_dx = - pdf * (dmean_dx + dstandard_devidation_dx * u)

        return cdf, dcdf_dx

    @property
    def has_gradients(self):
        return isinstance(self.model, IDifferentiable)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from scipy.special import erf
from typing import List, Tuple, Optional

from .quadrature_kernels import QuadratureKernel
from ...quadrature.interfaces.standard_kernels import IRBF
from ...quadrature.kernels.integration_measures import IntegrationMeasure, IsotropicGaussianMeasure, UniformMeasure
from ...quadrature.kernels.bounds import BoxBounds


class QuadratureRBF(QuadratureKernel):
    """
    Augments an RBF kernel with integrability

    Note 1: each standard kernel goes with a corresponding quadrature kernel, in this case the standard rbf kernel.
    Note 2: each child of this class implements a unique measure-integralBounds pair
    """

    def __init__(self, rbf_kernel: IRBF, integral_bounds: Optional[List[Tuple[float, float]]],
                 measure: Optional[IntegrationMeasure], variable_names: str='') -> None:
        """
        :param rbf_kernel: standard emukit rbf-kernel
        :param integral_bounds: defines the domain of the integral. List of D tuples, where D is the dimensionality
        of the integral and the tuples contain the lower and upper bounds of the integral
        i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]. None for infinite bounds
        :param measure: an integration measure. None means the standard Lebesgue measure is used.
        :param variable_names: the (variable) name(s) of the integral
        """

        super().__init__(kern=rbf_kernel, integral_bounds=integral_bounds, measure=measure, variable_names=variable_names)

    @property
    def lengthscale(self):
        return self.kern.lengthscale

    @property
    def variance(self):
        return self.kern.variance

    def qK(self, x2: np.ndarray) -> np.ndarray:
        """
        RBF kernel with the first component integrated out aka kernel mean

        :param x2: remaining argument of the once integrated kernel, shape (n_point N, input_dim)
        :returns: kernel mean at location x2, shape (1, N)
        """
        raise NotImplementedError

    def Kq(self, x1: np.ndarray) -> np.ndarray:
        """
        RBF kernel with the second component integrated out aka kernel mean

        :param x1: remaining argument of the once integrated kernel, shape (n_point N, input_dim)
        :returns: kernel mean at location x1, shape (N, 1)
        """
        return self.qK(x1).T

    def qKq(self) -> float:
        """
        RBF kernel integrated over both arguments x1 and x2

        :returns: double integrated kernel
        """
        raise NotImplementedError

    def dqK_dx(self, x2: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel mean (integrated in first argument) evaluated at x2
        :param x2: points at which to evaluate, shape (n_point N, input_dim)
        :return: the gradient with shape (input_dim, N)
        """
        raise NotImplementedError

    def dKq_dx(self, x1: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel mean (integrated in second argument) evaluated at x1
        :param x1: points at which to evaluate, shape (n_point N, input_dim)
        :return: the gradient with shape (N, input_dim)
        """
        return self.dqK_dx(x1).T

    # rbf-kernel specific helper
    def _scaled_vector_diff(self, v1: np.ndarray, v2: np.ndarray, scale: float=None) -> np.ndarray:
        """
        Scaled element-wise vector difference between vectors v1 and v2

        .. math::
            \frac{v_1 - v_2}{\lambda \sqrt{2}}

        name mapping:
            \lambda: self.kern.lengthscale

        :param v1: first vector
        :param v2: second vector, must have same second dimensions as v1
        :param scale: the scale, default is the lengthscale of the kernel
        :return: scaled difference between v1 and v2, np.ndarray with unchanged dimensions

        """
        if scale is None:
            scale = self.lengthscale
        return (v1 - v2) / (scale * np.sqrt(2))


class QuadratureRBFLebesgueMeasure(QuadratureRBF):
    """
    And RBF kernel with integrability over the standard Lebesgue measure. Can only be used with finite integral bounds.

    Note that each standard kernel goes with a corresponding quadrature kernel, in this case standard rbf kernel.
    """

    def __init__(self, rbf_kernel: IRBF, integral_bounds: List[Tuple[float, float]], variable_names: str='') -> None:
        """
        :param rbf_kernel: standard emukit rbf-kernel
        :param integral_bounds: defines the domain of the integral. List of D tuples, where D is the dimensionality
        of the integral and the tuples contain the lower and upper bounds of the integral
        i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]
        :param variable_names: the (variable) name(s) of the integral
        """
        super().__init__(rbf_kernel=rbf_kernel, integral_bounds=integral_bounds, measure=None,
                         variable_names=variable_names)

    def qK(self, x2: np.ndarray) -> np.ndarray:
        """
        RBF kernel with the first component integrated out aka kernel mean

        :param x2: remaining argument of the once integrated kernel, shape (n_point N, input_dim)
        :returns: kernel mean at location x2, shape (1, N)
        """
        lower_bounds = self.integral_bounds.lower_bounds
        upper_bounds = self.integral_bounds.upper_bounds
        erf_lo = erf(self._scaled_vector_diff(lower_bounds, x2))
        erf_up = erf(self._scaled_vector_diff(upper_bounds, x2))
        kernel_mean = self.variance * (self.lengthscale * np.sqrt(np.pi / 2.) * (erf_up - erf_lo)).prod(axis=1)

        return kernel_mean.reshape(1, -1)

    def qKq(self) -> float:
        """
        RBF kernel integrated over both arguments x1 and x2

        :returns: double integrated kernel
        """
        lower_bounds = self.integral_bounds.lower_bounds
        upper_bounds = self.integral_bounds.upper_bounds
        prefac = self.variance * (2. * self.lengthscale**2)**self.input_dim
        diff_bounds_scaled = self._scaled_vector_diff(upper_bounds, lower_bounds)
        exp_term = np.exp(-diff_bounds_scaled**2) - 1.
        erf_term = erf(diff_bounds_scaled) * diff_bounds_scaled * np.sqrt(np.pi)

        return float(prefac * (exp_term + erf_term).prod())

    def dqK_dx(self, x2: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel mean (integrated in first argument) evaluated at x2
        :param x2: points at which to evaluate, shape (n_point N, input_dim)
        :return: the gradient with shape (input_dim, N)
        """
        lower_bounds = self.integral_bounds.lower_bounds
        upper_bounds = self.integral_bounds.upper_bounds
        exp_lo = np.exp(- self._scaled_vector_diff(x2, lower_bounds) ** 2)
        exp_up = np.exp(- self._scaled_vector_diff(x2, upper_bounds) ** 2)
        erf_lo = erf(self._scaled_vector_diff(lower_bounds, x2))
        erf_up = erf(self._scaled_vector_diff(upper_bounds, x2))

        fraction = ((exp_lo - exp_up) / (self.lengthscale * np.sqrt(np.pi / 2.) * (erf_up - erf_lo))).T

        return self.qK(x2) * fraction


class QuadratureRBFIsoGaussMeasure(QuadratureRBF):
    """
    Augments an RBF kernel with integrability

    Note that each standard kernel goes with a corresponding quadrature kernel, in this case standard rbf kernel.
    """

    def __init__(self, rbf_kernel: IRBF, measure: IsotropicGaussianMeasure, variable_names: str='') -> None:
        """
        :param rbf_kernel: standard emukit rbf-kernel
        :param measure: a Gaussian measure
        :param variable_names: the (variable) name(s) of the integral
        """
        super().__init__(rbf_kernel=rbf_kernel, integral_bounds=None, measure=measure, variable_names=variable_names)

    def qK(self, x2: np.ndarray) -> np.ndarray:
        """
        RBF kernel with the first component integrated out aka kernel mean

        :param x2: remaining argument of the once integrated kernel, shape (n_point N, input_dim)
        :returns: kernel mean at location x2, shape (1, N)
        """
        det_factor = (self.measure.variance / self.lengthscale ** 2 + 1) ** (self.input_dim / 2)
        scale = np.sqrt(self.lengthscale ** 2 + self.measure.variance)
        scaled_vector_diff = self._scaled_vector_diff(x2, self.measure.mean, scale)
        kernel_mean = (self.variance / det_factor) * np.exp(- np.sum(scaled_vector_diff ** 2, axis=1))
        return kernel_mean.reshape(1, -1)

    def qKq(self) -> float:
        """
        RBF kernel integrated over both arguments x1 and x2

        :returns: double integrated kernel
        """
        factor = (2 * self.measure.variance / self.lengthscale ** 2 + 1) ** (self.input_dim / 2)
        result = self.variance / factor
        return float(result)

    def dqK_dx(self, x2: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel mean (integrated in first argument) evaluated at x2

        :param x2: points at which to evaluate, shape (n_point N, input_dim)
        :return: the gradient with shape (input_dim, N)
        """
        qK_x = self.qK(x2)
        factor = 1. / (self.lengthscale ** 2 + self.measure.variance)
        return - (qK_x * factor) * (x2 - self.measure.mean).T


class QuadratureRBFUniformMeasure(QuadratureRBF):
    """
    And RBF kernel with integrability over a uniform measure. Can be used with finite as well as infinite integral
    bounds.

    Note that each standard kernel goes with a corresponding quadrature kernel, in this case standard rbf kernel.
    """

    def __init__(self, rbf_kernel: IRBF, integral_bounds: Optional[List[Tuple[float, float]]],
                 measure: UniformMeasure, variable_names: str=''):
        """
        :param rbf_kernel: standard emukit rbf-kernel
        :param integral_bounds: defines the domain of the integral. List of D tuples, where D is the dimensionality
        of the integral and the tuples contain the lower and upper bounds of the integral
        i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]. None means infinite bounds.
        :param measure: A D-dimensional uniform measure
        :param variable_names: the (variable) name(s) of the integral
        """
        super().__init__(rbf_kernel=rbf_kernel, integral_bounds=integral_bounds, measure=measure,
                         variable_names=variable_names)

        # construct bounds that are used in the computation of the kernel integrals. The lower bounds are the max of
        # the lower bounds of integral and measure. The upper bounds are the min of the upper bounds of integral and
        # measure, i.e., the resulting bounds are the overlap over the integral bounds and the measure bounds.
        if integral_bounds is None:
            bounds = measure.get_box()
        else:
            bounds = [(max(ib[0], mb[0]), min(ib[1], mb[1])) for (ib, mb) in zip(integral_bounds, measure.get_box())]

        # checks if lower bounds are smaller than upper bounds.
        for (lb_d, ub_d) in bounds:
            if lb_d >= ub_d:
                raise ValueError("Upper bound of relevant integration domain must be larger than lower bound. Found a "
                                 "pair containing ({}, {}).".format(lb_d, ub_d))
        self._bounds_list_for_kernel_integrals = bounds
        self.reasonable_box_bounds = BoxBounds(name=variable_names, bounds=bounds)

    def qK(self, x2: np.ndarray) -> np.ndarray:
        """
        RBF kernel with the first component integrated out aka kernel mean

        :param x2: remaining argument of the once integrated kernel, shape (n_point N, input_dim)
        :returns: kernel mean at location x2, shape (1, N)
        """
        lower_bounds = np.array([b[0] for b in self._bounds_list_for_kernel_integrals])
        upper_bounds = np.array([b[1] for b in self._bounds_list_for_kernel_integrals])
        erf_lo = erf(self._scaled_vector_diff(lower_bounds, x2))
        erf_up = erf(self._scaled_vector_diff(upper_bounds, x2))
        kernel_mean = self.variance * (self.lengthscale * np.sqrt(np.pi / 2.) * (erf_up - erf_lo)).prod(axis=1)

        return kernel_mean.reshape(1, -1) * self.measure.density

    def qKq(self) -> float:
        """
        RBF kernel integrated over both arguments x1 and x2

        :returns: double integrated kernel
        """
        lower_bounds = np.array([b[0] for b in self._bounds_list_for_kernel_integrals])
        upper_bounds = np.array([b[1] for b in self._bounds_list_for_kernel_integrals])
        prefac = self.variance * (2. * self.lengthscale**2)**self.input_dim
        diff_bounds_scaled = self._scaled_vector_diff(upper_bounds, lower_bounds)
        exp_term = np.exp(-diff_bounds_scaled**2) - 1.
        erf_term = erf(diff_bounds_scaled) * diff_bounds_scaled * np.sqrt(np.pi)

        return float(prefac * (exp_term + erf_term).prod()) * self.measure.density ** 2

    def dqK_dx(self, x2: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel mean (integrated in first argument) evaluated at x2
        :param x2: points at which to evaluate, shape (n_point N, input_dim)
        :return: the gradient with shape (input_dim, N)
        """
        lower_bounds = np.array([b[0] for b in self._bounds_list_for_kernel_integrals])
        upper_bounds = np.array([b[1] for b in self._bounds_list_for_kernel_integrals])
        exp_lo = np.exp(- self._scaled_vector_diff(x2, lower_bounds) ** 2)
        exp_up = np.exp(- self._scaled_vector_diff(x2, upper_bounds) ** 2)
        erf_lo = erf(self._scaled_vector_diff(lower_bounds, x2))
        erf_up = erf(self._scaled_vector_diff(upper_bounds, x2))

        fraction = ((exp_lo - exp_up) / (self.lengthscale * np.sqrt(np.pi / 2.) * (erf_up - erf_lo))).T

        return self.qK(x2) * fraction



from time import time

import numpy as np

from ...core.loop import LoopState, OuterLoop


class Metric:
    def evaluate(self, loop: OuterLoop, loop_state: LoopState) -> None:
        pass

    def reset(self) -> None:
        pass


class MeanSquaredErrorMetric(Metric):
    """
    Mean squared error metric stored in loop state metric dictionary with key "mean_squared_error".
    """

    def __init__(self, x_test: np.ndarray, y_test: np.ndarray, name: str='mean_squared_error'):
        """
        :param x_test: Input locations of test data
        :param y_test: Test targets
        """
        self.x_test = x_test
        self.y_test = y_test
        self.name = name

    def evaluate(self, loop: OuterLoop, loop_state: LoopState) -> np.ndarray:
        """
        Calculate and store mean squared error

        :param loop: Outer loop
        :param loop_state: Object containing history of the loop that we add results to
        """
        # Calculate mean squared error
        predictions = loop.model_updaters[0].model.predict(self.x_test)[0]
        mse = np.mean(np.square(self.y_test - predictions), axis=0)
        return mse


class MinimumObservedValueMetric(Metric):
    """
    The result is stored in the "metrics" dictionary in the loop state with the key "minimum_observed_value"
    """

    def __init__(self, name: str='minimum_observed_value'):
        self.name = name

    def evaluate(self, loop: OuterLoop, loop_state: LoopState) -> np.ndarray:
        """
        Evaluates minimum observed value

        :param loop: Outer loop
        :param loop_state: Object containing history of the loop that we add results to
        """
        y_min = np.min(loop_state.Y, axis=0)
        return y_min


class TimeMetric(Metric):
    """
    Time taken between each iteration of the loop
    """

    def __init__(self, name: str='time'):
        """
        :param name: Name of the metric. Defaults to "time"
        """
        self.start_time = None
        self.name = name

    def evaluate(self, loop: OuterLoop, loop_state: LoopState) -> np.ndarray:
        """
        Returns difference between time now and when the reset method was last called
        """
        time_since_start = time() - self.start_time
        # Add to metrics dictionary in loop state
        return np.array([time_since_start])

    def reset(self) -> None:
        """
        Resets the start time
        """
        self.start_time = time()


class CumulativeCostMetric(Metric):
    """

    Accumulates the cost of each function evaluation.  The result is stored in the "metrics"
    dictionary in the loop state with the key "cumulative_costs"
    """

    def __init__(self, name: str='cumulative_costs'):
        self.name = name
        self.cumulative_costs = np.array([0.0])
        self.last_observed_iter = 0

    def evaluate(self, loop: OuterLoop, loop_state: LoopState) -> np.ndarray:
        """
        Computes the cumulative cost of all function evaluations after the last observed iteration

        :param loop: Outer loop
        :param loop_state: Object containing history of the loop that we add results to

        :return cumulative cost
        """
        if loop_state.cost[-1] is not None:
            self.cumulative_costs += np.cumsum(loop_state.cost[self.last_observed_iter:])[-1]
            self.last_observed_iter = loop_state.cost.shape[0]

        return self.cumulative_costs

    def reset(self) -> None:
        """
        Resets the cumulative cost and the internal counter back to 0
        """
        self.last_observed_iter = 0
        self.cumulative_costs = np.array([0.0])
